
/*
 * Linux device driver for ADMtek ADM8211 (IEEE 802.11b MAC/BBP)
 *
 * Copyright (c) 2003, Jouni Malinen <jkmaline@cc.hut.fi>
 * Copyright (c) 2004, Michael Wu <flamingice@sourmilk.net>
 * Some parts copyright (c) 2003 by David Young <dyoung@pobox.com>
 * and used with permission.
 *
 * Much thanks to Infineon-ADMtek for their support of this driver.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation. See README and COPYING for
 * more details.
 */

/*
 * Note!
 * - this is a pre-release of the driver; it is not yet finished and needs
 *   more testing. it should mostly work for infrastructure modes
 *
 * FIX:
 * - protect access of CSR_NAR with a spinlock
 * - properly idle chip in parts that need it (goes w/ above)
 * - improve signal quality + RSSI stuff
 * - figure out why ping -s 5913 and up lose packets..
 *   (can be reduced with a mtu of 920..)
 *
 * TODO:
 * - Determine optimum copybreak size?
 */


#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/if.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/ioport.h>
#include <linux/pci.h>
#include <linux/proc_fs.h>
#include <linux/if_arp.h>
#include <linux/delay.h>
#include <linux/crc32.h>
#include <linux/wireless.h>
#if WIRELESS_EXT > 12
#include <net/iw_handler.h>
#endif /* WIRELESS_EXT > 12 */
#include <asm/delay.h>
#include <asm/unaligned.h>
#include <asm/types.h>
#include <asm/div64.h>

#undef ADM8211_PROC
#include "adm8211.h"
#include "adm8211_ioctl.h"
#include "ieee80211.h"


MODULE_AUTHOR("Jouni Malinen <jkmaline@cc.hut.fi>, Michael Wu <flamingice@sourmilk.net>");
MODULE_DESCRIPTION("Driver for IEEE 802.11b wireless LAN cards based on ADMtek"
		   " ADM8211");
MODULE_SUPPORTED_DEVICE("ADM8211");
MODULE_LICENSE("GPL");


static const char *version = KERN_INFO "adm8211: Copyright 2003, "
"Jouni Malinen <jkmaline@cc.hut.fi>; "
"Copyright 2004, Michael Wu <flamingice@sourmilk.net>\n";


static struct pci_device_id adm8211_pci_id_table[] __devinitdata = {
	/* ADMtek ADM8211 */
	{ 0x1200, 0x8201, PCI_ANY_ID, PCI_ANY_ID },
	{ 0x1317, 0x8201, PCI_ANY_ID, PCI_ANY_ID },
	{ 0x1317, 0x8211, PCI_ANY_ID, PCI_ANY_ID },
	{ 0 }
};

#ifdef ADM8211_PROC
static struct proc_dir_entry *adm8211_proc;
#endif


#define ADM8211_INTMASK \
(ADM8211_IER_NIE | ADM8211_IER_AIE | ADM8211_IER_RCIE | ADM8211_IER_TCIE | \
ADM8211_IER_TDUIE | ADM8211_IER_GPTIE)





struct adm8211_desc {
	u32 status;
	u32 length;
	u32 buffer1;
	u32 buffer2;
};

#define RDES0_STATUS_OWN	(1 << 31)
#define RDES0_STATUS_ES		(1 << 30)
#define RDES0_STATUS_SQL	(1 << 29)
#define RDES0_STATUS_DE		(1 << 28)
#define RDES0_STATUS_FS		(1 << 27)
#define RDES0_STATUS_LS		(1 << 26)
#define RDES0_STATUS_PCF	(1 << 25)
#define RDES0_STATUS_SFDE	(1 << 24)
#define RDES0_STATUS_SIGE	(1 << 23)
#define RDES0_STATUS_CRC16E	(1 << 22)
#define RDES0_STATUS_RXTOE	(1 << 21)
#define RDES0_STATUS_CRC32E	(1 << 20)
#define RDES0_STATUS_ICVE	(1 << 19)
#define RDES0_STATUS_DA1	(1 << 17)
#define RDES0_STATUS_DA0	(1 << 16)
#define RDES0_STATUS_RXDR	((1 << 15) | (1 << 14) | (1 << 13) | (1 << 12))
#define RDES0_STATUS_FL		(0x00000fff)

#define RDES1_CONTROL_RER	(1 << 25)
#define RDES1_CONTROL_RCH	(1 << 24)
#define RDES1_CONTROL_RBS2	(0x00fff000)
#define RDES1_CONTROL_RBS1	(0x00000fff)

#define RDES1_STATUS_RSSI	(0x0000007f)


#define TDES0_CONTROL_OWN	(1 << 31)
#define TDES0_CONTROL_DONE	(1 << 30)
#define TDES0_CONTROL_TXDR	(0x0ff00000)

#define TDES0_STATUS_OWN	(1 << 31)
#define TDES0_STATUS_DONE	(1 << 30)
#define TDES0_STATUS_ES		(1 << 29)
#define TDES0_STATUS_TLT	(1 << 28)
#define TDES0_STATUS_TRT	(1 << 27)
#define TDES0_STATUS_TUF	(1 << 26)
#define TDES0_STATUS_TRO	(1 << 25)
#define TDES0_STATUS_SOFBR	(1 << 24)
#define TDES0_STATUS_ACR	(0x00000fff)

#define TDES1_CONTROL_IC	(1 << 31)
#define TDES1_CONTROL_LS	(1 << 30)
#define TDES1_CONTROL_FS	(1 << 29)
#define TDES1_CONTROL_TER	(1 << 25)
#define TDES1_CONTROL_TCH	(1 << 24)
#define TDES1_CONTROL_RBS2	(0x00fff000)
#define TDES1_CONTROL_RBS1	(0x00000fff)


#define PLCP_SIGNAL_1M		0x0a
#define PLCP_SIGNAL_2M		0x14
#define PLCP_SIGNAL_5M5		0x37
#define PLCP_SIGNAL_11M		0x6e


/* RX status - stored in skb->cb so this structure must be 48 bytes or less */
struct adm8211_rx_status {
	u8 rssi;
	u8 rate;
};


struct adm8211_tx_hdr {
	u8 da[6];
	u8 signal; /* PLCP signal / TX rate in 100 Kbps */
	u8 service;
	u16 frame_body_size;
	u16 frame_control;
	u16 plcp_frag_tail_len;
	u16 plcp_frag_head_len;
	u16 dur_frag_tail;
	u16 dur_frag_head;
	u8 address4[6];
	u16 header_control;
#ifdef BIG_ENDIAN
	u32 retry_limit:8;
	u32 frag_number:4;
	u32 reserved_0:8;
	u32 frag_threshold:12;
#else
	u32 frag_threshold:12;
	u32 reserved_0:8;
	u32 frag_number:4;
	u32 retry_limit:8;
#endif

	u32 wep2key0;
	u32 wep2key1;
	u32 wep2key2;
	u32 wep2key3;

	u8 keyid;
	u8 reserved_1[3];
	u32 reserved_2;
} __attribute__ ((packed));



#define RX_COPY_BREAK 256

#define RX_PKT_SIZE 2400



/* SRAM offsets */
#define ADM8211_SRAM_INDIV_KEY 0x0000
#define ADM8211_SRAM_SHARE_KEY 0x0160
#define ADM8211_SRAM_SSID      0x0180
#define ADM8211_SRAM_SUPP_RATE 0x0191
#define ADM8211_SRAM_SIZE      0x0200

static void adm8211_write_sram(struct net_device *dev, int addr, u16 data);
static int adm8211_write_bbp(struct net_device *dev, u8 addr, u8 data);
static int adm8211_tx(struct net_device *dev, struct sk_buff *skb);


/* Serial EEPROM reading for 93C66/93C46 */
#define EE_ENB		(0x4000 | ADM8211_SPR_SRS | ADM8211_SPR_SCS)
#define EE_READ_CMD	(6)
#define eeprom_delay()	ADM8211_CSR_READ(SPR);


static u16 adm8211_eeprom_read_word(struct net_device *dev, int addr,
				    int addr_len)
{
	struct adm8211_priv *priv = dev->priv;
	int i;
	u16 retval = 0;
	int read_cmd = addr | (EE_READ_CMD << addr_len);

	ADM8211_CSR_WRITE(SPR, EE_ENB & ~ADM8211_SPR_SCS);
	ADM8211_CSR_WRITE(SPR, EE_ENB);

	/* Shift the read command bits out. */
	for (i = 4 + addr_len; i >= 0; i--) {
		short dataval = (read_cmd & (1 << i)) ? ADM8211_SPR_SDI : 0;
		ADM8211_CSR_WRITE(SPR, EE_ENB | dataval);
		eeprom_delay();
		ADM8211_CSR_WRITE(SPR, EE_ENB | dataval | ADM8211_SPR_SCLK);
		eeprom_delay();
	}

	ADM8211_CSR_WRITE(SPR, EE_ENB);
	eeprom_delay();

	for (i = 16; i > 0; i--) {
		ADM8211_CSR_WRITE(SPR, EE_ENB | ADM8211_SPR_SCLK);
		eeprom_delay();
		retval <<= 1;
		retval |= ((ADM8211_CSR_READ(SPR) & ADM8211_SPR_SDO) ? 1 : 0);
		ADM8211_CSR_WRITE(SPR, EE_ENB);
		eeprom_delay();
	}

	/* Terminate the EEPROM access. */
	ADM8211_CSR_WRITE(SPR, EE_ENB & ~ADM8211_SPR_SCS);

	return retval;
}


static int adm8211_read_eeprom(struct net_device *dev)
{
	struct adm8211_priv *priv = dev->priv;
	u32 reg;
	int addr_len, words, i;

	reg = ADM8211_CSR_READ(CSR_TEST0);

	if (reg & ADM8211_CSR_TEST0_EPNE) {
		printk(KERN_INFO "%s: EEPROM not detected\n", dev->name);
		goto guess;
	}

	if (reg & ADM8211_CSR_TEST0_EPTYP) {
		printk(KERN_DEBUG "%s: EEPROM type: 93C66\n", dev->name);
		/* 256 * 16-bit = 512 bytes */
		addr_len = 8;
		words = 256;
	} else {
		printk(KERN_DEBUG "%s: EEPROM type: 93C46\n", dev->name);
		/* 64 * 16-bit = 128 bytes */
		addr_len = 6;
		words = 64;
	}

	priv->eeprom_len = words * 2;
	priv->eeprom = kmalloc(priv->eeprom_len, GFP_KERNEL);
	if (priv->eeprom == NULL)
		return -ENOMEM;

	for (i = 0; i < words; i++)
		*((u16 *) &priv->eeprom[i * 2]) =
			adm8211_eeprom_read_word(dev, i, addr_len);

	if (reg & ADM8211_CSR_TEST0_EPSNM) {
		printk(KERN_INFO "%s: EEPROM signature not matched\n",
		       dev->name);
		goto guess;
	}

	priv->rf_type = (priv->eeprom[0x16] & 0x0038) >> 3;
	priv->bbp_type = priv->eeprom[0x16] & 0x0007;

	printk(KERN_DEBUG "%s: RFtype=%d BBPtype=%d\n",
	       dev->name, priv->rf_type, priv->bbp_type);

	/* read country code from eeprom and make sure it's valid */
	i=(priv->eeprom[0x0D] & 0xFF00) >> 8;
	if (i >= ARRAY_SIZE(cranges)) {
		printk(KERN_INFO "%s: Invalid country code in EEPROM, assuming ETSI\n",
		       dev->name);
		priv->ieee80211.chan_range = cranges[2];
		return -1;
	}
	priv->ieee80211.chan_range = cranges[i];

	printk(KERN_DEBUG "%s: Channel range: %d-%d\n",
	       dev->name, (int)priv->ieee80211.chan_range.min, (int)priv->ieee80211.chan_range.max);
	return 0;

guess:
	printk(KERN_WARNING "%s: Guessing RF/BBP type and country code\n", dev->name);
	priv->rf_type = 1;
	priv->bbp_type = 1;
	priv->ieee80211.chan_range = cranges[2]; /* ETSI */

	return -1;
}

static void adm8211_write_sram(struct net_device *dev, int addr, u16 data)
{
	struct adm8211_priv *priv = dev->priv;
	u32 reg = ADM8211_CSR_READ(WEPCTL);

	ADM8211_CSR_WRITE(WEPCTL, cpu_to_le32(addr | ADM8211_WEPCTL_TABLE_WR));
	ADM8211_CSR_READ(WEPCTL);
	mdelay(1);
	ADM8211_CSR_WRITE(WESK, cpu_to_le32(data));
	ADM8211_CSR_WRITE(WEPCTL, reg);
}

static void adm8211_clear_sram(struct net_device *dev)
{
	int addr;

	for (addr = 0; addr < ADM8211_SRAM_SIZE; addr++)
		adm8211_write_sram(dev, addr, 0);
}

static void adm8211_link_change(struct net_device *dev, int link)
{
	struct adm8211_priv *priv = dev->priv;
	u32 reg;

#if 0
	/* FIX: this does not seem to be LinkLed ? */
	reg = ADM8211_CSR_READ(GPIO);
	reg &= ~(ADM8211_CSR_GPIO_IN);
	reg |= ADM8211_CSR_GPIO_EN0;
	if (link)
		reg |= ADM8211_CSR_GPIO_O0;
	else
		reg &= ~ADM8211_CSR_GPIO_O0;
	ADM8211_CSR_WRITE(GPIO, reg);
#endif

	/* Set aid for beacon TIM element decoding */
	reg = ADM8211_CSR_READ(FRCTL);
	reg &= ~ADM8211_FRCTL_AID;
	reg |= ADM8211_FRCTL_AID_ON;
	if (link)
		reg |= priv->ieee80211.aid;
	ADM8211_CSR_WRITE(FRCTL, reg);
	ADM8211_CSR_READ(FRCTL);
}

static struct net_device_stats *adm8211_get_stats(struct net_device *dev)
{
	return &((struct adm8211_priv *)dev->priv)->stats;
}

static struct iw_statistics *adm8211_get_wireless_stats(struct net_device *dev)
{
	struct adm8211_priv *priv = dev->priv;
	struct iw_statistics *wstats = &priv->wstats;
	return wstats;
}

static void adm8211_set_rx_mode (struct net_device *dev)
{
	struct adm8211_priv *priv = dev->priv;
	int i, bit_nr;
	u32 reg, mc_filter[2];
	struct dev_mc_list *mclist;

	reg = ADM8211_CSR_READ(NAR);
	/* make sure receive is stopped */
	if (reg & ADM8211_NAR_SR) {
	ADM8211_CSR_WRITE(NAR, reg & ~ADM8211_NAR_SR);
	mdelay(20);
	}

	if (dev->flags & IFF_PROMISC) {
		reg |= ADM8211_NAR_PR;
		reg &= ~ADM8211_NAR_MM;
		dev->flags &= ~IFF_ALLMULTI;
		mc_filter[1] = mc_filter[0] = 0xffffffff;
	} else if ((dev->flags & IFF_ALLMULTI) || (dev->mc_count > 32)) {
		reg &= ~ADM8211_NAR_PR;
		reg |= ADM8211_NAR_MM;
		dev->flags |= IFF_ALLMULTI;
		mc_filter[1] = mc_filter[0] = 0xffffffff;
	} else {
		reg &= ~(ADM8211_NAR_MM | ADM8211_NAR_PR);
		dev->flags &= ~IFF_ALLMULTI;
		/* Multicast filter calculation adapted from the 8139too driver */
		mc_filter[1] = mc_filter[0] = 0;
		mclist = dev->mc_list;
		for (i = 0; i < dev->mc_count; i+=1) {
			if (!mclist)
				break;
			bit_nr = ether_crc(ETH_ALEN, mclist->dmi_addr) >> 26;

                        mc_filter[bit_nr >> 5] |= 1 << (bit_nr & 31);
			mclist = mclist->next;
		}
	}

	ADM8211_CSR_WRITE(MAR0, mc_filter[0]);
	ADM8211_CSR_WRITE(MAR1, mc_filter[1]);
	ADM8211_CSR_WRITE(NAR, reg);
	ADM8211_CSR_READ(NAR);
}

static int adm8211_set_mac_address (struct net_device *dev, void *p)
{
	struct adm8211_priv *priv = dev->priv;
        struct sockaddr *addr = p;
	u32 reg;

        memcpy(dev->dev_addr, addr->sa_data, dev->addr_len);

	/* sure, we can change the MAC addr while running */
	if (dev->flags & IFF_UP) {
	reg = ADM8211_CSR_READ(NAR);

	/* make sure receive is stopped */
	if (reg & ADM8211_NAR_SR) {
	ADM8211_CSR_WRITE(NAR, reg & ~ADM8211_NAR_SR);
	mdelay(20);
	}

	/* FIX: ugly */
	ADM8211_CSR_WRITE(PAR0, (dev->dev_addr[0]<<0 ) + (dev->dev_addr[1]<<8 ) +
				(dev->dev_addr[2]<<16) + (dev->dev_addr[3]<<24) );
	ADM8211_CSR_WRITE(PAR1, (dev->dev_addr[4]<<0 ) + (dev->dev_addr[5]<<8 ) );
	ADM8211_CSR_WRITE(NAR, reg);
	}

	return 0;
}

static void adm8211_rx_skb(struct net_device *dev, struct sk_buff *skb)
{
	struct adm8211_priv *priv = dev->priv;
	struct ieee80211_hdr *hdr;
	u16 fc;
	struct adm8211_rx_status *stat;
	struct iw_statistics *wstats = &priv->wstats;
	struct ieee80211_rx_status rx_status;

	if (priv->iw_mode == IW_MODE_MONITOR) {
		memset(skb->cb, 0, sizeof(skb->cb));
		skb = ieee80211_data_decode(&priv->ieee80211, skb);
		skb->protocol = eth_type_trans(skb, dev);
		priv->stats.rx_bytes+=skb->len;
		priv->stats.rx_packets+=1;
		netif_rx(skb);
		return;
	}

	hdr = (struct ieee80211_hdr *) skb->data;
	stat = (struct adm8211_rx_status *) skb->cb;

	if (skb->len < 14)
		goto drop;

	fc = le16_to_cpu(hdr->frame_control);

	if (ieee80211_filter_duplicates(&priv->ieee80211, skb, fc))
		goto drop;

	/* remove FCS */
	if (dev->flags & IFF_PROMISC)
		skb_trim(skb, skb->len - FCS_LEN);

	/* FIX: this is a hack */
	wstats->qual.qual=stat->rssi;
	wstats->qual.updated=1;
	switch (WLAN_FC_GET_TYPE(fc)) {
	case WLAN_FC_TYPE_MGMT:
		memset(&rx_status, 0, sizeof(rx_status));
		rx_status.rssi = stat->rssi;
		switch (stat->rate) {
		case 0:
			rx_status.rate = 10;
			break;
		case 1:
			rx_status.rate = 20;
			break;
		case 2:
			rx_status.rate = 55;
			break;
		case 3:
			rx_status.rate = 110;
			break;
		case 4:
			rx_status.rate = 220;
			break;
		}
		ieee80211_rx_mgmt(&priv->ieee80211, skb, &rx_status);
		break;
	case WLAN_FC_TYPE_DATA:
		skb = ieee80211_data_decaps(&priv->ieee80211, skb);
		if (skb) {
			skb->protocol = eth_type_trans(skb, dev);
			memset(skb->cb, 0, sizeof(skb->cb));
			priv->stats.rx_bytes+=skb->len;
			priv->stats.rx_packets+=1;
			netif_rx(skb);
		}
		break;
	case WLAN_FC_TYPE_CTRL:
/*		printk(KERN_DEBUG "RX ctrl, type %d - dropped\n", (int)WLAN_FC_GET_STYPE(fc));*/
		goto drop;
	}

	return;
 drop:
	dev_kfree_skb(skb);
}


static void adm8211_rx_tasklet(unsigned long data)
{
	struct net_device *dev = (struct net_device *) data;
	struct adm8211_priv *priv = dev->priv;
	struct sk_buff *skb;

	while ((skb = skb_dequeue(&priv->rx_queue)) != NULL)
		adm8211_rx_skb(dev, skb);
}


static void adm8211_interrupt_tci(struct net_device *dev)
{
	struct adm8211_priv *priv = dev->priv;
	unsigned dirty_tx;

#if 0
	printk(KERN_DEBUG "TCI (dirty_tx=%d cur_tx=%d)\n",
	       priv->dirty_tx, priv->cur_tx);
#endif

	spin_lock(&priv->lock);

	for (dirty_tx = priv->dirty_tx;
	     priv->cur_tx - dirty_tx > 0; dirty_tx++) {
		unsigned entry = dirty_tx % TX_RING_SIZE;
		u32 status = le32_to_cpu(priv->tx_ring[entry].status);
		if (status & TDES0_CONTROL_OWN ||
		    !(status & TDES0_CONTROL_DONE))
			break;

		if (status & TDES0_STATUS_ES) {
		priv->stats.tx_errors+=1;

		if (status & (TDES0_STATUS_TLT | TDES0_STATUS_SOFBR))
			priv->wstats.discard.misc+=1;
		if (status & (TDES0_STATUS_TUF | TDES0_STATUS_TRO))
			priv->stats.tx_fifo_errors+=1;
		if (status & TDES0_STATUS_TRT)
			priv->wstats.discard.retries+=1;
		} else {
		priv->stats.tx_bytes+=priv->tx_buffers[entry].skb->len;
		priv->stats.tx_packets+=1;
		}

		pci_unmap_single(priv->pdev, priv->tx_buffers[entry].mapping,
				 priv->tx_buffers[entry].skb->len,
				 PCI_DMA_TODEVICE);
		dev_kfree_skb_irq(priv->tx_buffers[entry].skb);
		priv->tx_buffers[entry].skb = NULL;
	}

	if (priv->cur_tx - dirty_tx < TX_RING_SIZE - 2)
		netif_wake_queue(dev);

	priv->dirty_tx = dirty_tx;
	spin_unlock(&priv->lock);
}


static void adm8211_interrupt_rci(struct net_device *dev)
{
	struct adm8211_priv *priv = dev->priv;
	int entry = priv->cur_rx % RX_RING_SIZE;
	u32 status;
	unsigned pktlen;
	struct sk_buff *skb, *newskb;
	int limit = RX_RING_SIZE;
	u8 rssi, rate;

#if 0
	printk(KERN_DEBUG "RCI\n");
#endif

	while (!(priv->rx_ring[entry].status &
		 __constant_cpu_to_le32(RDES0_STATUS_OWN))) {
		if (--limit < 0)
			break;

		status = le32_to_cpu(priv->rx_ring[entry].status);
		rate = (status & RDES0_STATUS_RXDR) >> 12;
		rssi = le32_to_cpu(priv->rx_ring[entry].length) &
			RDES1_STATUS_RSSI;

#if 0
		printk(KERN_DEBUG "%s: RX %02x RXDR=%d FL=%d RSSI=%d "
		       "%s%s%s%s\n",
		       dev->name, status, rate,
		       status & RDES0_STATUS_FL,
		       rssi,
		       status & RDES0_STATUS_ES ? "[ES]" : "",
		       status & RDES0_STATUS_FS ? "[FS]" : "",
		       status & RDES0_STATUS_LS ? "[LS]" : "",
		       status & RDES0_STATUS_CRC16E ? "[CRC16E]" : "");
#endif

		pktlen = status & RDES0_STATUS_FL;
		if (pktlen > RX_PKT_SIZE) {
			printk(KERN_DEBUG "%s: too long frame (pktlen=%d)\n",
			       dev->name, pktlen);
			pktlen = RX_PKT_SIZE;
		}

		if (status & RDES0_STATUS_ES) {
#if 0
			printk(KERN_DEBUG "%s: dropped RX frame with error "
			       "(status=0x%x)\n", dev->name, status);
#endif
			skb = NULL; /* old buffer will be reused */
			priv->stats.rx_errors+=1;
			if (status & (RDES0_STATUS_PCF | RDES0_STATUS_SFDE |
				      RDES0_STATUS_SIGE | RDES0_STATUS_RXTOE))
				priv->wstats.discard.misc+=1;
			if (status & RDES0_STATUS_ICVE)
				priv->wstats.discard.code+=1;
			if (status & (RDES0_STATUS_CRC16E | RDES0_STATUS_CRC32E))
				priv->stats.rx_crc_errors+=1;

		} else if (pktlen < RX_COPY_BREAK) {
			skb = dev_alloc_skb(pktlen);
			if (skb) {
				skb->dev = dev;
				pci_dma_sync_single(
					priv->pdev,
					priv->rx_buffers[entry].mapping,
					pktlen, PCI_DMA_FROMDEVICE);
				memcpy(skb_put(skb, pktlen),
				       priv->rx_buffers[entry].skb->tail,
				       pktlen);
			}
		} else {
			newskb = dev_alloc_skb(RX_PKT_SIZE);
			if (newskb) {
				newskb->dev = dev;
				skb = priv->rx_buffers[entry].skb;
				skb_put(skb, pktlen);
				pci_unmap_single(
					priv->pdev,
					priv->rx_buffers[entry].mapping,
					RX_PKT_SIZE, PCI_DMA_FROMDEVICE);
				priv->rx_buffers[entry].skb = newskb;
				priv->rx_buffers[entry].mapping =
					pci_map_single(priv->pdev,
						       newskb->tail,
						       RX_PKT_SIZE,
						       PCI_DMA_FROMDEVICE);
			} else {
				skb = NULL;
				priv->stats.rx_dropped+=1;
			}

			priv->rx_ring[entry].buffer1 =
				cpu_to_le32(priv->rx_buffers[entry].mapping);
		}

		priv->rx_ring[entry].status = cpu_to_le32( RDES0_STATUS_OWN | RDES0_STATUS_SQL | RDES0_STATUS_FS | RDES0_STATUS_LS );
		priv->rx_ring[entry].length =
			cpu_to_le32(RX_PKT_SIZE |
				    (entry == RX_RING_SIZE - 1 ?
				     RDES1_CONTROL_RER : 0));

		if (skb) {
			struct adm8211_rx_status *stat =
				(struct adm8211_rx_status *) skb->cb;
#if 0
			{
				int i;
				printk(KERN_DEBUG "RX[%d/%d]",
				       pktlen, skb->len);
				for (i = 0; i < skb->len; i++)
					printk(" %02x", skb->data[i]);
				printk("\n");
			}
#endif
			stat->rssi = rssi;
			stat->rate = rate;
			skb->mac.raw = skb->data;
			skb->protocol = __constant_htons(ETH_P_802_2);
			skb_queue_tail(&priv->rx_queue, skb);
			tasklet_schedule(&priv->rx_tasklet);
			dev->last_rx = jiffies;
		}

		entry = (++priv->cur_rx) % RX_RING_SIZE;
	}

	priv->stats.rx_missed_errors+=ADM8211_CSR_READ(LPC) & 0xFFFF;
}


static irqreturn_t adm8211_interrupt(int irq, void *dev_id,
				     struct pt_regs *regs)
{
	struct net_device *dev = dev_id;
	struct adm8211_priv *priv = dev->priv;
	u32 stsr;
	int count = 0;

	do {
		stsr = ADM8211_CSR_READ(STSR);
		ADM8211_CSR_WRITE(STSR, stsr);
		if (stsr == 0xffffffff)
			return IRQ_HANDLED;

		if (!(stsr & (ADM8211_STSR_NISS | ADM8211_STSR_AISS)))
			break;

		if (stsr & ADM8211_STSR_PCF)
			printk(KERN_DEBUG "PCF\n");
		if (stsr & ADM8211_STSR_BCNTC)
			printk(KERN_DEBUG "BCNTC\n");
		if (stsr & ADM8211_STSR_GPINT)
			printk(KERN_DEBUG "GPINT\n");
		if (stsr & ADM8211_STSR_LinkOff) {
			printk(KERN_DEBUG "LinkOff\n");
			/* HACK */
			if (dev->flags & IFF_UP)
				ieee80211_start(&priv->ieee80211);
		}
		if (stsr & ADM8211_STSR_ATIMTC)
			printk(KERN_DEBUG "ATIMTC\n");
		if (stsr & ADM8211_STSR_TSFTF)
			printk(KERN_DEBUG "TSFTF\n");
		if (stsr & ADM8211_STSR_TSCZ)
			printk(KERN_DEBUG "TSCZ\n");
		if (stsr & ADM8211_STSR_LinkOn)
			printk(KERN_DEBUG "LinkOn\n");
		if (stsr & ADM8211_STSR_SQL)
			printk(KERN_DEBUG "SQL\n");
		if (stsr & ADM8211_STSR_WEPTD)
			printk(KERN_DEBUG "WEPTD\n");
		if (stsr & ADM8211_STSR_ATIME)
			printk(KERN_DEBUG "ATIME\n");
		if (stsr & ADM8211_STSR_TBTT)
			printk(KERN_DEBUG "TBTT\n");
		if (stsr & ADM8211_STSR_TEIS)
			printk(KERN_DEBUG "TEIS\n");
		if (stsr & ADM8211_STSR_FBE)
			printk(KERN_DEBUG "FBE\n");
		if (stsr & ADM8211_STSR_REIS)
			printk(KERN_DEBUG "REIS\n");
		if (stsr & ADM8211_STSR_GPTT) {
			printk(KERN_DEBUG "GPTT\n");
			/* ADM8211_CSR_WRITE(GPTMR, 10 * 4902); */
		}
		if (stsr & ADM8211_STSR_RPS)
			printk(KERN_DEBUG "RPS\n");
		if (stsr & ADM8211_STSR_RDU)
			printk(KERN_DEBUG "RDU\n");
		if (stsr & ADM8211_STSR_RCI)
			adm8211_interrupt_rci(dev);
		if (stsr & ADM8211_STSR_TUF)
			printk(KERN_DEBUG "TUF\n");
		if (stsr & ADM8211_STSR_TRT)
			priv->wstats.discard.retries+=1;
/*		if (stsr & ADM8211_STSR_TLT)
			printk(KERN_DEBUG "TLT\n");*/
		if (stsr & ADM8211_STSR_TDU) {
			/*printk(KERN_DEBUG "TDU\n");*/
		}
		if (stsr & ADM8211_STSR_TPS)
			printk(KERN_DEBUG "TPS\n");
		if (stsr & ADM8211_STSR_TCI)
			adm8211_interrupt_tci(dev);

		/* TODO: is this needed with non-Cardbus cards? */
		ADM8211_CSR_WRITE(FER, ADM8211_FER_INTR_EV_ENT);

	} while (count++ < 20);

	/* Testing, clear forced interrupt */
	ADM8211_CSR_WRITE(FFER, 0);

	return IRQ_RETVAL(count);
}


#if 0
static int adm8211_rf_read_syn(struct net_device *dev, u32 addr)
{
	struct adm8211_priv *priv = dev->priv;
	int timeout = 10;
	u32 reg;

	ADM8211_CSR_WRITE(SYNCTL, ADM8211_SYNCTL_RD | ADM8211_SYNCTL_CS0 |
			  ADM8211_SYNCTL_RFDM | addr);

	while (timeout > 0) {
		mdelay(2);
		reg = ADM8211_CSR_READ(SYNCTL);
		if (!(reg & ADM8211_SYNCTL_RD))
			break;
		timeout--;
	}

	if (timeout == 0) {
		printk(KERN_DEBUG "%s: SYNCTL timeout(addr=0x%x) reg=0x%x\n",
		       dev->name, addr, reg);
		return -1;
	}

	return (reg & 0x3fffff) >> 4;
}
#endif /* 0 */


static int adm8211_rf_write_syn(struct net_device *dev, u32 addr, u32 data)
{
	struct adm8211_priv *priv = dev->priv;
	int timeout;

	timeout=20;
	while (timeout > 0) {
		if (!(ADM8211_CSR_READ(SYNCTL) & (ADM8211_SYNCTL_WR | ADM8211_SYNCTL_RD)))
			break;
		timeout--;
		mdelay(1);
	}
	if (timeout == 0) {
		printk(KERN_DEBUG "%s: SYNCTL prewrite timeout(addr=0x%x, data=0x%x)\n",
		       dev->name, addr, data);
		return -1;
	}

	ADM8211_CSR_WRITE(SYNCTL, ADM8211_SYNCTL_WR | ADM8211_SYNCTL_CS0 |
			  ADM8211_SYNCTL_RFDM | data << 4 | addr);

	timeout=20;
	while (timeout > 0) {
		if (!(ADM8211_CSR_READ(SYNCTL) & ADM8211_SYNCTL_WR))
			break;
		timeout--;
		mdelay(1);
	}

	ADM8211_CSR_WRITE(SYNCTL, ADM8211_CSR_READ(SYNCTL) & (~ADM8211_SYNCTL_WR));
	if (timeout == 0) {
		printk(KERN_DEBUG "%s: SYNCTL postwrite timeout(addr=0x%x, data=0x%x)\n",
		       dev->name, addr, data);
		return -1;
	}

	return 0;
}


int adm8211_rf_set_channel(struct net_device *dev, int channel)
{
	struct adm8211_priv *priv = dev->priv;
	u32 old, reg;

	if (channel < 1 || channel > 14)
		return -1;

	ADM8211_CSR_WRITE(SYNRF, ADM8211_SYNRF_SELRF | ADM8211_SYNRF_PE1 |
			  ADM8211_SYNRF_PHYRST);
	ADM8211_CSR_READ(SYNRF);

	mdelay(20);

	/* Stop transmit */
	old = reg = ADM8211_CSR_READ(NAR);
	reg &= ~(ADM8211_NAR_ST | ADM8211_NAR_SR);
	ADM8211_CSR_WRITE(NAR, reg);
	ADM8211_CSR_READ(NAR);

	mdelay(20);
#if 0
	/* Program Si4126 RF Synthesizer to the new channel */
	if (adm8211_rf_write_syn(dev, SI4126_MAIN_CONF, 0)) {
		/* First write seems to timeout for some reason.. retry */
		adm8211_rf_write_syn(dev, SI4126_MAIN_CONF, 0);
	}
	adm8211_rf_write_syn(dev, SI4126_PHASE_DET_GAIN, 0);
	adm8211_rf_write_syn(dev, SI4126_POWERDOWN,
			     SI4126_POWERDOWN_PDIB | SI4126_POWERDOWN_PDRB);
	adm8211_rf_write_syn(dev, SI4126_RF2_N_DIV, 2038 + (channel - 1) * 5);
	adm8211_rf_write_syn(dev, SI4126_IF_N_DIV, 1496);
	adm8211_rf_write_syn(dev, SI4126_RF2_R_DIV, 88);
	adm8211_rf_write_syn(dev, SI4126_IF_R_DIV, 88);
#else
	/* Program Si4126 RF Synthesizer to the new channel */
	if (adm8211_rf_write_syn(dev, SI4126_MAIN_CONF, SI4126_MAIN_XINDIV2)) {
		/* First write seems to timeout for some reason.. retry */
		adm8211_rf_write_syn(dev, SI4126_MAIN_CONF,
				     SI4126_MAIN_XINDIV2);
	}
	adm8211_rf_write_syn(dev, SI4126_PHASE_DET_GAIN, 0);
	adm8211_rf_write_syn(dev, SI4126_POWERDOWN,
			     SI4126_POWERDOWN_PDIB | SI4126_POWERDOWN_PDRB);
	/* TODO: check if this works for channel 14 */
	adm8211_rf_write_syn(dev, SI4126_RF2_N_DIV, 2038 + (channel - 1) * 5);
	adm8211_rf_write_syn(dev, SI4126_IF_N_DIV, 1496);
	adm8211_rf_write_syn(dev, SI4126_RF2_R_DIV, 44);
	adm8211_rf_write_syn(dev, SI4126_IF_R_DIV, 44);
#endif


#if 1
	/* write BBP regs */
	if (priv->bbp_type == ADM8211_TYPE_RFMD) {
		/* RF3000 BBP */
		adm8211_write_bbp(dev, 0x11, 0xc8);
		adm8211_write_bbp(dev, 0x14, 0x14);
		adm8211_write_bbp(dev, 0x15,
				  (channel==14) ? 0x50 : 0xd0);
		adm8211_write_bbp(dev, 0x1c, 0x00);
		adm8211_write_bbp(dev, 0x1d, 0x84);
		/* ? */
		adm8211_write_bbp(dev, 0x11, 0xbc);
	} else {
		printk(KERN_DEBUG "%s: unsupport BBP type %d\n",
		       dev->name, priv->bbp_type);
	}
#endif

#if 0
	/* Restart transmit */
	reg = ADM8211_CSR_READ(NAR);
	reg |= ADM8211_NAR_ST | ADM8211_NAR_SR;
	ADM8211_CSR_WRITE(NAR, reg);
#else
	ADM8211_CSR_WRITE(NAR, old);
#endif

	ADM8211_CSR_WRITE(SYNRF, 0);

	ADM8211_CSR_WRITE(CAP0, channel);
#if 0
	reg = ADM8211_CSR_READ(GPIO);
	reg &= 0xfffc000;
	reg |= ADM8211_CSR_GPIO_EN0;
	if (0)
		reg |= ADM8211_CSR_GPIO_O0;
	ADM8211_CSR_WRITE(GPIO, reg);
#endif

	return 0;
}

void adm8211_write_wepkey(struct net_device *dev, int index)
{
#define ADM8211_WEP_ENABLE	(1 << 7)
#define ADM8211_WEP_104		(1 << 6)

	struct adm8211_priv *priv = dev->priv;
	struct ieee80211_data *data = &priv->ieee80211;
	int i, addr;
	u16 key;

	addr = (index*7) + ADM8211_SRAM_SHARE_KEY;
	if (data->wep_data[index].len > 5)
		key = (ADM8211_WEP_ENABLE | ADM8211_WEP_104) << 8;
	else if (data->wep_data[index].len != 0)
		key = ADM8211_WEP_ENABLE << 8;
	else
		key = 0;

	adm8211_write_sram(dev, addr, data->wep_data[index].key[0] | key);

	/* len is never even, but we don't care about the remainder */
	for (i = 1; i <= (data->wep_data[index].len/2); i+=1)
		adm8211_write_sram(dev, addr + i,
				   data->wep_data[index].key[i*2] +
				   (data->wep_data[index].key[i*2+1] << 8));
}

void adm8211_update_wep(struct net_device *dev)
{
/*	struct adm8211_priv *priv = dev->priv;
	struct ieee80211_data *data = &priv->ieee80211;
	u32 reg;*/
/*
	reg = ADM8211_CSR_READ(MACTEST);
	if (data->wep_enabled) {
		reg |= (1<<22);
		reg &= ~(3<<20);
		reg |= (data->default_key << 20);
	} else
		reg &= ~(7<<20);
	ADM8211_CSR_WRITE(MACTEST, reg);

	reg = ADM8211_CSR_READ(WEPCTL);
	if (data->wep_enabled)
		reg |= ADM8211_WEPCTL_WEPENABLE;
	else
		reg &= ~ADM8211_WEPCTL_WEPENABLE;

	if (priv->revid < ADM8211_REV_BA)
		reg |= ADM8211_WEPCTL_WEPRXBYP;
	else
		reg &= ~ADM8211_WEPCTL_WEPRXBYP;
	ADM8211_CSR_WRITE(WEPCTL, reg);
	ADM8211_CSR_READ(WEPCTL);*/
}

/* TODO: set ATIMW in CAP1? */
static void adm8211_update_mode(struct net_device *dev)
{
	struct adm8211_priv *priv = dev->priv;
	struct ieee80211_data *data = &priv->ieee80211;
	u32 reg;
	unsigned int len, rem;

	switch (priv->iw_mode) {
		case IW_MODE_INFRA:
			dev->type = ARPHRD_ETHER;
			dev->hard_header_parse = priv->eth_header_parse;
			reg = ADM8211_CSR_READ(NAR);
			reg &= ~(ADM8211_NAR_PR | ADM8211_NAR_EA);
			ADM8211_CSR_WRITE(NAR, reg);
			ADM8211_CSR_WRITE(BCNT, 0);
			ADM8211_CSR_WRITE(CAP1, 0);
			ADM8211_CSR_READ(NAR);
			break;
		case IW_MODE_ADHOC:
			dev->type = ARPHRD_ETHER;
			dev->hard_header_parse = priv->eth_header_parse;
			reg = ADM8211_CSR_READ(NAR);
			reg &= ~ADM8211_NAR_PR;
			reg |= ADM8211_NAR_EA;
			ADM8211_CSR_WRITE(NAR, reg);

			/* FIX */
			len = 26 +
			      6 + 2 + 2 + 2 + data->ssid_len + 2 + data->num_supp_rates +
			      3 + 4 + WLAN_FCS_LEN;
			len *= 10;
			rem  = len % priv->plcp_signal;
			len /= priv->plcp_signal;
			if (rem)
				len+=1;
			reg = len;
			reg &= 0x7FFF;
			/* set bit 15 if tx rate is 11M */
			if (data->rate == 0x16)
				reg |= 1<<15;
			ADM8211_CSR_WRITE(BCNT, reg);
			ADM8211_CSR_WRITE(CAP1, data->capab<<16);
			ADM8211_CSR_READ(NAR);
			break;
		case IW_MODE_MONITOR:
			dev->type = ARPHRD_IEEE80211;
			dev->hard_header_parse = adm8211_80211_header_parse;
			reg = ADM8211_CSR_READ(NAR);
			reg &= ~ADM8211_NAR_EA;
			reg |= ADM8211_NAR_PR;
			ADM8211_CSR_WRITE(NAR, reg);
			ADM8211_CSR_WRITE(BCNT, 0);
			ADM8211_CSR_WRITE(CAP1, 0);
			ADM8211_CSR_READ(NAR);
			break;
	}
}

static int adm8211_write_bbp(struct net_device *dev, u8 addr, u8 data)
{
	struct adm8211_priv *priv = dev->priv;
	int timeout;
	u32 reg;

	timeout = 20;
	while (timeout > 0) {
		reg = ADM8211_CSR_READ(BBPCTL);
		if (!(reg & (ADM8211_BBPCTL_WR | ADM8211_BBPCTL_RD)))
			break;
		timeout--;
		mdelay(1);
	}
	if (timeout == 0) {
		printk(KERN_DEBUG "%s: adm8211_write_bbp(%d,%d) failed prewrite "
		       "(reg=0x%08x)\n",
		       dev->name, addr, data, reg);
		return -ETIMEDOUT;
	}

	reg = (0x20<<24) | (1<<23) | (1<<21) | (0x1<<18) | ADM8211_BBPCTL_WR;
	reg |= addr << 8;
	reg |= data;

	ADM8211_CSR_WRITE(BBPCTL, reg);

	timeout=20;
	while (timeout > 0) {
		reg = ADM8211_CSR_READ(BBPCTL);
		if (!(reg & ADM8211_BBPCTL_WR))
			break;
		timeout--;
		mdelay(1);
	}

	ADM8211_CSR_WRITE(BBPCTL, ADM8211_CSR_READ(BBPCTL) & (~ADM8211_BBPCTL_WR));

	if (timeout == 0) {
		/* FIX: WR bit is supposed to be cleared, but for some reason
		 * this does not seem to happen.. Could just comment out this
		 * warning(?) */
		printk(KERN_DEBUG "%s: adm8211_write_bbp(%d,%d) failed postwrite "
		       "(reg=0x%08x)\n",
		       dev->name, addr, data, reg);
		return -ETIMEDOUT;
	}

	return 0;
}


static int adm8211_hw_init_bbp(struct net_device *dev)
{
	struct adm8211_priv *priv = dev->priv;
	u32 reg, old;

	/* Set RF Power control IF pin to PE1+PHYRST# */
	ADM8211_CSR_WRITE(SYNRF, 0x42040000);
	ADM8211_CSR_READ(SYNRF);
	mdelay(20);

	/* Stop transmit */
	old = reg = ADM8211_CSR_READ(NAR);
	reg &= ~(ADM8211_NAR_ST | ADM8211_NAR_SR);
	ADM8211_CSR_WRITE(NAR, reg);
	ADM8211_CSR_READ(NAR);

	/* write BBP regs */
	if (priv->bbp_type == ADM8211_TYPE_RFMD) {
		/* RF3000 BBP */
		/* another set:
		 * 11: c8
		 * 14: 14
		 * 15: 50 (chan 1..13; chan 14: d0)
		 * 1c: 00
		 * 1d: 84
		 */
		adm8211_write_bbp(dev, 0x03, 0x01<<7);
		adm8211_write_bbp(dev, 0x11, 0x74);
		adm8211_write_bbp(dev, 0x14, 0x38);
		adm8211_write_bbp(dev, 0x15, 0x40);
		adm8211_write_bbp(dev, 0x1c, 0x00);
		adm8211_write_bbp(dev, 0x1d, 0x80);
		adm8211_write_bbp(dev, 0x02, 0x01<<6);
		/* ? */
		adm8211_write_bbp(dev, 0x11, 0xbc);
	} else {
		printk(KERN_DEBUG "%s: unsupport BBP type %d\n",
		       dev->name, priv->bbp_type);
	}

#if 0
	reg = ADM8211_CSR_READ(NAR);
	/* FIX: start transmit? */
	ADM8211_CSR_WRITE(NAR, reg);
#else
	ADM8211_CSR_WRITE(NAR, old);
#endif

	/* Set RF CAL control source to MAC control */
	reg = ADM8211_CSR_READ(SYNCTL);
	reg |= ADM8211_SYNCTL_SELCAL;
	ADM8211_CSR_WRITE(SYNCTL, reg);

	return 0;
}

static int adm8211_set_rate(struct net_device *dev)
{
	struct adm8211_priv *priv = dev->priv;
	struct ieee80211_data *data = &priv->ieee80211;
	struct ieee80211_bss *bss;
	u32 reg;

	/* keep bits 0-14 */
	reg = ADM8211_CSR_READ(PLCPHD) & 0x7FFF;

	/* Service: 0xb0 */
	reg |= 0x00b00000;

	spin_lock_bh(&data->lock);
	bss = ieee80211_get_bss(data, data->bssid);

	if (bss && (bss->capability & WLAN_CAPABILITY_SHORT_PREAMBLE))
		reg |= (1 << 15);

	spin_unlock_bh(&data->lock);

	priv->plcp_signal = priv->ieee80211.rate*5;

	reg |= priv->plcp_signal << 24;
	ADM8211_CSR_WRITE(PLCPHD, reg);

	/* MTMLT   = 512 TU (max TX MSDU lifetime)
	 * BCNTSIG = plcp_signal (beacon, probe resp, and atim TX rate)
	 * SRTYLIM = 224 (short retry limit) */
	ADM8211_CSR_WRITE(TXLMT, (512<<16) + (priv->plcp_signal<<7) + (224<<0));

	ADM8211_CSR_READ(TXLMT);
	return 0;
}

static void adm8211_hw_init(struct net_device *dev)
{
	struct adm8211_priv *priv = dev->priv;
	u32 reg;
	u8 cacheline;

	/* FIX: ? */
	/*ADM8211_CSR_WRITE(CSR_TEST1, 0x18100);*/
	/*ADM8211_CSR_WRITE(WCSR, 0xe00021);*/
	reg = (0x07 << 21) | (1 << 20) | (1 << 8);
	ADM8211_CSR_WRITE(WCSR, reg);

	/* Disable APM, enable receive FIFO threshold, and set drain receive
	 * threshold to store-and-forward */
	reg = ADM8211_CSR_READ(CMDR);
	reg &= ~(ADM8211_CMDR_APM | ADM8211_CMDR_DRT);
	reg |= ADM8211_CMDR_RTE | (0x2 << 2);
	ADM8211_CSR_WRITE(CMDR, reg);

	adm8211_set_rate(dev);

	/* 4-bit values:
	 * PWR1UP = 8 * 2 ms
	 * PWR0PAPE = 8 us or 5 us
	 * PWR1PAPE = 1 us or 3 us
	 * PWR0TRSW = 5 us
	 * PWR1TRSW = 12 us
	 * PWR0PE2 = 13 us
	 * PWR1PE2 = 1 us
	 * PWR0TXPE = 8 or 6 */
	if (priv->revid < ADM8211_REV_CA)
		ADM8211_CSR_WRITE(TOFS2, 0x8815cd18);
	else
		ADM8211_CSR_WRITE(TOFS2, 0x8535cd16);

	/* Enable store and forward for transmit */
	ADM8211_CSR_WRITE(NAR, ADM8211_NAR_SF | ADM8211_NAR_PB);

	/* Set bit 20 (reset RF) in SYNRF */
	ADM8211_CSR_WRITE(SYNRF, 0x100000);
	ADM8211_CSR_READ(SYNRF);
	mdelay(20);
	ADM8211_CSR_WRITE(SYNRF, 0);
	ADM8211_CSR_READ(SYNRF);

	mdelay(10);

	/* Set CFP Max Duration to 0x10 TU */
	reg = ADM8211_CSR_READ(CFPP);
	reg &= 0xff0000ff;
	reg |= 0x00001000;
	ADM8211_CSR_WRITE(CFPP, reg);


	/* USCNT = 0x16 (number of system clocks, 22 MHz, in 1us
	 * TUCNT = 0x3ff - Tu counter 1024 us  */
	ADM8211_CSR_WRITE(TOFS0, (0x16 << 24) | 0x3ff);

	/* SLOT=20 us, SIFS=220 cycles of 22 MHz (10 us),
	 * DIFS=50 us, EIFS=100 us */
	ADM8211_CSR_WRITE(IFST, (20 << 23) | (220 << 15) | (50 << 9) | 100);

	/* PCNT = 1 (MAC idle time awake/sleep, unit S)
	 * RMRD = 18769 us (max RX duration)  */
	ADM8211_CSR_WRITE(RMD, (1 << 16) | 18769);

	/* MART=65535 us, MIRT=256 us, TSFTOFST=0 us */
	ADM8211_CSR_WRITE(RSPT, 0xffffff00);

	/* FIX: ?? */
	ADM8211_CSR_WRITE(MMIWA,  0x00009101);
	ADM8211_CSR_WRITE(MMIRD0, 0x00000301);
	ADM8211_CSR_WRITE(MMIRD1, 0x7e100000);
	ADM8211_CSR_WRITE(MACTEST, 0x800);

	/* Initialize BBP */
	adm8211_hw_init_bbp(dev);

	/* ACK interrupts */
	ADM8211_CSR_WRITE(STSR, ADM8211_CSR_READ(STSR));

	/* Supported Rates (for ad hoc mode) */
	adm8211_write_sram(dev, ADM8211_SRAM_SUPP_RATE, 0);
	adm8211_write_sram(dev, ADM8211_SRAM_SUPP_RATE + 1, 0x8204);
	adm8211_write_sram(dev, ADM8211_SRAM_SUPP_RATE + 2, 0x8b84);
	adm8211_write_sram(dev, ADM8211_SRAM_SUPP_RATE + 3, 0x0096);
	adm8211_write_sram(dev, ADM8211_SRAM_SUPP_RATE + 4, 0x0000);
	adm8211_write_sram(dev, ADM8211_SRAM_SUPP_RATE + 5, 0x5800);

	reg = ADM8211_CSR_READ(PAR);
	reg |= (0x1<<23) | (0x1<<21);
	reg &= ~((0x1<<1) | (0x3<<14));

	/* FIX */
/*	if (!pci_set_mwi(priv->pdev))
		reg |= (0x1<<24);*/

	pci_read_config_byte(priv->pdev, PCI_CACHE_LINE_SIZE, &cacheline);
	switch (cacheline) {
		case  0x8: reg |= (0x1<<14);
			   break;
		case 0x16: reg |= (0x2<<14);
			   break;
		case 0x32: reg |= (0x3<<14);
			   break;
		  default: reg |= (0x0<<14);
			   break;
	}

	ADM8211_CSR_WRITE(PAR, reg);
}


static int adm8211_hw_reset(struct net_device *dev)
{
	struct adm8211_priv *priv = dev->priv;
	u32 reg;
	int ret = 0;

	/* Reset the chip */
	reg = ADM8211_CSR_READ(PAR);
	ADM8211_CSR_WRITE(PAR, ADM8211_PAR_SWR);

	while ((ADM8211_CSR_READ(PAR) & ADM8211_PAR_SWR) && ret < 40) {
	udelay(50);
	ret+=1;
	}
	if (ret>=40) {
		printk(KERN_DEBUG "%s: software reset timed out\n", dev->name);
		ret = -ETIMEDOUT;
	}
	ADM8211_CSR_WRITE(PAR, reg);
	ADM8211_CSR_READ(PAR);

	ADM8211_CSR_WRITE(FRCTL, 0);

	/* is this check necessary? */
	if (priv->revid < ADM8211_REV_BA)
		adm8211_clear_sram(dev);

	return ret;
}

/* First beacon was sent at time 0 microseconds, current time is
 * tsfth << 32 | tsftl microseconds, and beacon interval is tbtt
 * microseconds.  Return the expected time in microseconds for the
 * beacon after next.
 */
static inline uint64_t adm8211_predict_beacon(uint64_t __tsft, uint32_t __tbtt)
{
	uint64_t tmp = __tsft;
	return __tsft + (__tbtt - do_div(tmp, __tbtt));
}

static void adm8211_beacon_update(struct net_device *dev, uint16_t duration)
{
	struct adm8211_priv *priv = dev->priv;
	struct ieee80211_data *data = &priv->ieee80211;
	struct ieee80211_bss *bss;
	uint32_t tsftl;
	uint64_t tsft, tbtt;

	spin_lock_bh(&data->lock);
	bss = ieee80211_get_bss(data, data->bssid);
	if (!bss)
		return;

	if (data->mode != IEEE80211_ADHOC) {
		tbtt=adm8211_predict_beacon(bss->timestamp, bss->interval << 10);
	} else {
		/* If we read TSFTL right before rollover, we read a TSF timer
		 * that is too high rather than too low. This prevents a spurious
		 * synchronization down the line, however, our IBSS could suffer
		 * from a creeping TSF....
		 */
		tsftl = ADM8211_CSR_READ(TSFTL);
		tsft = ((uint64_t)ADM8211_CSR_READ(TSFTH) << 32 | tsftl);

		/* we are faster, let the other guy catch up */
		if (bss->timestamp < tsft) {
			spin_unlock_bh(&data->lock);
			return;
		}

		tbtt=adm8211_predict_beacon(tsft, bss->interval << 10);
	}
	spin_unlock_bh(&data->lock);

	/* convert to TUs */
	tbtt >>= 10;

	/* wake up a little early */
	tbtt -= 20;

	/* TSFTOFSR (RX TSFT Offset) = 1 us
	 * TBTTPRE (prediction time) = tbtt TU
	 * TBTTOFS (Wake up time offset before TBTT) = 20 TU */
	ADM8211_CSR_WRITE(TOFS1, (30 << 24) | ((uint32_t)tbtt << 8) | 20);
	ADM8211_CSR_READ(TOFS1);
}

static void adm8211_auth_update(struct net_device *dev)
{
	struct adm8211_priv *priv = dev->priv;
	struct ieee80211_data *data = &priv->ieee80211;
	struct ieee80211_bss *bss;
	u32 reg;

	adm8211_set_rate(dev);
	adm8211_beacon_update(dev, 0);

	spin_lock_bh(&data->lock);
	bss = ieee80211_get_bss(data, data->bssid);
	/* BP (beacon interval) = bss->interval
	 * LI (listen interval) = data->listen_interval/bss->interval (beacon intervals) */
	ADM8211_CSR_WRITE(BPLI, (bss->interval << 16) || (data->listen_interval/bss->interval));
	spin_unlock_bh(&data->lock);

	/* lose link after 2 lost beacons */
	reg = ADM8211_CSR_READ(WCSR) & ~(0xFF<<21);
	ADM8211_CSR_WRITE(WCSR, reg & (0x02<<21));
	ADM8211_CSR_READ(WCSR);
}

static int adm8211_set_bssid(struct net_device *dev, u8 *bssid)
{
	struct adm8211_priv *priv = dev->priv;
	u32 reg;

	reg = bssid[0] | (bssid[1] << 8) | (bssid[2] << 16) | (bssid[3] << 24);
	ADM8211_CSR_WRITE(BSSID0, reg);
	reg = ADM8211_CSR_READ(ABDA1);
	reg &= 0x0000ffff;
	reg |= (bssid[4] << 16) | (bssid[5] << 24);
	ADM8211_CSR_WRITE(ABDA1, reg);

	if (bssid[0] != 0xff)
		adm8211_beacon_update(dev, 0);

	return 0;
}


static int adm8211_set_ssid(struct net_device *dev, u8 *ssid, size_t ssid_len)
{
	u8 buf[34];
	int i;

	if (ssid_len > 32)
		return -1;
	memset(buf, 0, sizeof(buf));
	buf[0] = ssid_len;
	memcpy(buf + 1, ssid, ssid_len);
	for (i = 0; i < sizeof(buf); i += 2) {
		u16 val = buf[i] | buf[i + 1] << 8;
		adm8211_write_sram(dev, ADM8211_SRAM_SSID + i / 2, val);
	}

	return 0;
}


static int adm8211_open(struct net_device *dev)
{
	struct adm8211_priv *priv = dev->priv;
	int retval;
	u32 reg;

	/* Power up MAC and RF chips */
	/*ADM8211_CSR_WRITE(FRCTL, 0);*/

	adm8211_hw_reset(dev);

/*	ADM8211_CSR_WRITE(FRCTL, 0);

	adm8211_clear_sram(dev);*/

	/* init ring */
	ADM8211_CSR_WRITE(RDB, priv->rx_ring_dma);
	ADM8211_CSR_WRITE(TDBD, priv->tx_ring_dma);
	priv->cur_rx = priv->cur_tx = priv->dirty_tx = 0;

	/* Init hardware */
	adm8211_hw_init(dev);

	udelay(200);
	adm8211_rf_set_channel(dev, 1);
	adm8211_update_mode(dev);
	adm8211_set_rx_mode(dev);
	adm8211_write_wepkey(dev, 0);
	adm8211_write_wepkey(dev, 1);
	adm8211_write_wepkey(dev, 2);
	adm8211_write_wepkey(dev, 3);
	adm8211_update_wep(dev);

	retval = request_irq(dev->irq, &adm8211_interrupt, SA_SHIRQ, dev->name,
			     dev);
	if (retval) {
		printk(KERN_ERR "%s: failed to register IRQ handler\n",
		       dev->name);
		return retval;
	}

	ADM8211_CSR_WRITE(IER, ADM8211_INTMASK);

	/* set mac address */
        ADM8211_CSR_WRITE(PAR0, (dev->dev_addr[0]<<0 ) + (dev->dev_addr[1]<<8 ) +
				(dev->dev_addr[2]<<16) + (dev->dev_addr[3]<<24) );
        ADM8211_CSR_WRITE(PAR1, (dev->dev_addr[4]<<0 ) + (dev->dev_addr[5]<<8 ) );

	reg = ADM8211_CSR_READ(NAR);
	reg |= ADM8211_NAR_ST | ADM8211_NAR_SR;
	ADM8211_CSR_WRITE(NAR, reg);
	ADM8211_CSR_READ(NAR);

	ieee80211_start(&priv->ieee80211);

	netif_start_queue(dev);

	return 0;
}


static int adm8211_stop(struct net_device *dev)
{
	struct adm8211_priv *priv = dev->priv;
	u32 reg;

	spin_lock(&priv->lock);
	ieee80211_stop(&priv->ieee80211);

	netif_stop_queue(dev);

	ADM8211_CSR_WRITE(IER, 0);
	ADM8211_CSR_READ(IER);
	reg = ADM8211_CSR_READ(NAR);
	reg &= ~(ADM8211_NAR_ST | ADM8211_NAR_PR | ADM8211_NAR_PB |
		 ADM8211_NAR_SR | ADM8211_NAR_MM );
	ADM8211_CSR_WRITE(NAR, reg);
	ADM8211_CSR_READ(NAR);

	free_irq(dev->irq, dev);
	spin_unlock(&priv->lock);

	adm8211_hw_reset(dev);
	return 0;
}

static void adm8211_calc_durations(int *dur, int *plcp, size_t payload_len, int len,
				   int plcp_signal, int short_preamble)
{
#if 0
	int tmp;

	/*
	 * IEEE 802.11b, Ch. 18.3.4
	 * aSIFSTime = 10 usec
	 * aPreambleLength = 144 usec (72 usec with short preamble)
	 * aPLCPHeaderLength = 48 usec (24 usec with short preamble)
	 */

	switch (plcp_signal) {
	case PLCP_SIGNAL_1M:
		plcp = len * 8;
		dur = plcp + 542;
		break;
	case PLCP_SIGNAL_2M:
		plcp = len * 4;
		dur = plcp + 542;
		break;
	case PLCP_SIGNAL_5M5:
		plcp = (len * 8 * 2) / 11;
		tmp = (len * 8 * 2) % 11;
		if (tmp)
			plcp+=1;
		dur = plcp + 430;
		break;
	case PLCP_SIGNAL_11M:
		plcp = (len * 8) / 11;
		tmp = (len * 8) % 11;
		if (tmp)
			plcp+=1;
		dur = plcp + 430;
		/*
		if (tmp == 3 || tmp == 7 || tmp == 10)
			plcp |= 0x8000;
		*/
		break;
	default:
		dev_kfree_skb(skb);
		return 0;
	}

	if (!short_preamble)
		dur += 288;
#else
	int remainder;

	/* Alternative calculation from NetBSD: */

/* IEEE 802.11b durations for DSSS PHY in microseconds */
#define IEEE80211_DUR_DS_LONG_PREAMBLE	144
#define IEEE80211_DUR_DS_SHORT_PREAMBLE	72
#define IEEE80211_DUR_DS_FAST_PLCPHDR	24
#define IEEE80211_DUR_DS_SLOW_PLCPHDR	48
#define IEEE80211_DUR_DS_SLOW_ACK	112
#define IEEE80211_DUR_DS_FAST_ACK	56
#define IEEE80211_DUR_DS_SLOW_CTS	112
#define IEEE80211_DUR_DS_FAST_CTS	56
#define IEEE80211_DUR_DS_SLOT		20
#define IEEE80211_DUR_DS_SIFS		10

	*dur = (80 * (WLAN_ADDR3_HDR_LEN + payload_len) + plcp_signal - 1)
		/ plcp_signal;

	if (plcp_signal <= PLCP_SIGNAL_2M) {
		/* 1-2Mbps WLAN: send ACK/CTS at 1Mbps */
		*dur += 3 * (IEEE80211_DUR_DS_SIFS +
			     IEEE80211_DUR_DS_SHORT_PREAMBLE +
			     IEEE80211_DUR_DS_FAST_PLCPHDR) +
			     IEEE80211_DUR_DS_SLOW_CTS + IEEE80211_DUR_DS_SLOW_ACK;
	} else {
		/* 5-11Mbps WLAN: send ACK/CTS at 2Mbps */
		*dur += 3 * (IEEE80211_DUR_DS_SIFS +
			     IEEE80211_DUR_DS_SHORT_PREAMBLE +
			     IEEE80211_DUR_DS_FAST_PLCPHDR) +
			     IEEE80211_DUR_DS_FAST_CTS + IEEE80211_DUR_DS_FAST_ACK;
	}

	/* lengthen duration if long preamble */
	if (!short_preamble) {
		*dur +=
			3 * (IEEE80211_DUR_DS_LONG_PREAMBLE -
			     IEEE80211_DUR_DS_SHORT_PREAMBLE) +
			3 * (IEEE80211_DUR_DS_SLOW_PLCPHDR -
			     IEEE80211_DUR_DS_FAST_PLCPHDR);
	}


	*plcp = (80 * len) / plcp_signal;
	remainder = (80 * len) % plcp_signal;
	if (plcp_signal == PLCP_SIGNAL_11M && remainder <= 30) {
		*plcp = (*plcp | 0x8000) + 1;
	} else if (remainder)
		(*plcp)++;
#endif
}

/* Transmit skb (with 802.11 header) */
static int adm8211_tx(struct net_device *dev, struct sk_buff *skb)
{
	struct adm8211_priv *priv = dev->priv;
	struct ieee80211_hdr *hdr;
	struct ieee80211_bss *bss;
	struct adm8211_tx_hdr *txhdr;
	int entry;
	dma_addr_t mapping;
	u32 flag;
	unsigned long flags;
	u16 fc, fc_;
	u8 da[ETH_ALEN];
	size_t payload_len;
	int plcp, dur, len;

	int plcp_signal;
	int short_preamble = 0;

	/* Move information from IEEE 802.11 header into ADM8211 TX header */
	hdr = (struct ieee80211_hdr *) skb->data;
	fc = hdr->frame_control;
	fc_ = le16_to_cpu(fc);
	if (WLAN_FC_GET_TYPE(fc_) == WLAN_FC_TYPE_DATA &&
	    (fc_ & (WLAN_FC_TODS | WLAN_FC_FROMDS)) == WLAN_FC_TODS)
		memcpy(da, hdr->addr3, ETH_ALEN);
	else
		memcpy(da, hdr->addr1, ETH_ALEN);
	skb_pull(skb, 24 /* FIX */);
	payload_len = skb->len;
	/*if (fc_ & WLAN_FC_ISWEP)
		payload_len += 8;*/

	if (skb_headroom(skb) < sizeof(struct adm8211_tx_hdr)) {
		if (pskb_expand_head(skb, sizeof(struct adm8211_tx_hdr), 0,
				     GFP_ATOMIC)) {
			printk(KERN_DEBUG "%s: failed to allocate room for TX "
			       "header\n", dev->name);
			dev_kfree_skb(skb);
			return 0;
		}
	}

	spin_lock_bh(&priv->ieee80211.lock);
	bss = ieee80211_get_bss(&priv->ieee80211, da);

	/* enable short preamble if we can */
	if (bss && (bss->capability & WLAN_CAPABILITY_SHORT_PREAMBLE))
		short_preamble = 1;
	spin_unlock_bh(&priv->ieee80211.lock);

	plcp_signal = ieee80211_get_rate(&priv->ieee80211, da)*5;

	txhdr = (struct adm8211_tx_hdr *) skb_push(skb, sizeof(*txhdr));
	memset(txhdr, 0, sizeof(*txhdr));
	memcpy(txhdr->da, da, ETH_ALEN);
	txhdr->signal = plcp_signal;
	txhdr->frame_body_size = cpu_to_le16(payload_len);
	txhdr->frame_control = fc;

	/* FIX: duration/len for fragmented frames */
#if 0
	len = min((WLAN_ADDR3_HDR_LEN + payload_len + WLAN_FCS_LEN), (priv->frag_thresh - 6));
	adm8211_calc_durations(&dur, &plcp, min((payload_len), (priv->frag_thresh - 34)),
				len, plcp_signal, short_preamble);
	txhdr->plcp_frag_head_len = cpu_to_le16(plcp);
	txhdr->dur_frag_head = cpu_to_le16(dur);

	len = WLAN_ADDR3_HDR_LEN + (payload_len%(priv->frag_thresh - 34)) + WLAN_FCS_LEN;
	adm8211_calc_durations(&dur, &plcp, (payload_len%(priv->frag_thresh - 34)),
				len, plcp_signal, short_preamble);
	txhdr->plcp_frag_tail_len = cpu_to_le16(plcp);
	txhdr->dur_frag_tail = cpu_to_le16(dur);
#else
	len = WLAN_ADDR3_HDR_LEN + payload_len + WLAN_FCS_LEN;
	adm8211_calc_durations(&dur, &plcp, payload_len, len, plcp_signal, short_preamble);
	txhdr->plcp_frag_tail_len = cpu_to_le16(plcp);
	txhdr->plcp_frag_head_len = cpu_to_le16(plcp);
	txhdr->dur_frag_tail = cpu_to_le16(dur);
	txhdr->dur_frag_head = cpu_to_le16(dur);
#endif

	txhdr->header_control = cpu_to_le16(short_preamble ? 0x8001 : 0x8000);
/*	if (fc_ & WLAN_FC_ISWEP) {
		txhdr->header_control |= cpu_to_le16(0x20);
	}*/
	txhdr->frag_threshold = cpu_to_le16(priv->frag_thresh & 0x0FFF);
	txhdr->retry_limit = 3;

	mapping = pci_map_single(priv->pdev, skb->data, skb->len,
				 PCI_DMA_TODEVICE);

	spin_lock_irqsave(&priv->lock, flags);

	entry = priv->cur_tx % TX_RING_SIZE;

	priv->tx_buffers[entry].skb = skb;
	priv->tx_buffers[entry].mapping = mapping;
	priv->tx_ring[entry].buffer1 = cpu_to_le32(mapping);

	if (priv->cur_tx - priv->dirty_tx < TX_RING_SIZE / 2)
		flag = TDES1_CONTROL_LS | TDES1_CONTROL_FS;
	else if (priv->cur_tx - priv->dirty_tx == TX_RING_SIZE / 2)
		flag = TDES1_CONTROL_IC | TDES1_CONTROL_LS | TDES1_CONTROL_FS;
	else if (priv->cur_tx - priv->dirty_tx < TX_RING_SIZE - 2)
		flag = TDES1_CONTROL_LS | TDES1_CONTROL_FS;
	else {
		flag = TDES1_CONTROL_IC | TDES1_CONTROL_LS | TDES1_CONTROL_FS;
		netif_stop_queue(dev);
	}
	if (entry == TX_RING_SIZE - 1)
		flag |= TDES1_CONTROL_TER;
	priv->tx_ring[entry].length = cpu_to_le32(flag | skb->len);

	/* Set TX rate (SIGNAL field in PLCP PPDU format) */
	flag = TDES0_CONTROL_OWN | (plcp_signal << 20) |
		3 /* retry count */;
	priv->tx_ring[entry].status = cpu_to_le32(flag);

	priv->cur_tx++;

	spin_unlock_irqrestore(&priv->lock, flags);

	/* Trigger transmit poll */
	ADM8211_CSR_WRITE(TDR, 0);
	ADM8211_CSR_READ(TDR);

	dev->trans_start = jiffies;

	return 0;
}


static int adm8211_hard_start_xmit(struct sk_buff *skb, struct net_device *dev)
{
	struct adm8211_priv *priv = dev->priv;

	skb = ieee80211_data_encaps(&priv->ieee80211, skb);
	if (skb)
		return adm8211_tx(dev, skb);
	return 0;
}


int adm8211_80211_header_parse(struct sk_buff *skb, unsigned char *haddr)
{
	memcpy(haddr, skb->mac.raw + 10, ETH_ALEN); /* addr2 */
	return ETH_ALEN;
}


static void adm8211_init_rings(struct net_device *dev)
{
	struct adm8211_priv *priv = dev->priv;
	int i;
	struct adm8211_desc *desc;
	struct adm8211_ring_info *info;

	for (i = 0; i < RX_RING_SIZE; i++) {
		desc = &priv->rx_ring[i];
		desc->status = 0;
		desc->length = cpu_to_le32(RX_PKT_SIZE);
		desc->buffer2 = cpu_to_le32(priv->rx_ring_dma + (i + 1) *
					    sizeof(struct adm8211_desc));
		priv->rx_buffers[i].skb = NULL;
	}
	/* Mark the end of RX ring; hw returns to base address after this
	 * descriptor */
	desc->length |= cpu_to_le32(RDES1_CONTROL_RER);
	/* Even though RER should indicate return to start, set next pointer
	 * to the base address just in case (some tulip cards require this). */
	desc->buffer2 = cpu_to_le32(priv->rx_ring_dma);


	for (i = 0; i < RX_RING_SIZE; i++) {
		desc = &priv->rx_ring[i];
		info = &priv->rx_buffers[i];

		info->skb = dev_alloc_skb(RX_PKT_SIZE);
		if (info->skb == NULL)
			break;
		info->mapping = pci_map_single(priv->pdev, info->skb->tail,
					       RX_PKT_SIZE,
					       PCI_DMA_FROMDEVICE);
		info->skb->dev = dev;
		desc->buffer1 = cpu_to_le32(info->mapping);
		desc->status = cpu_to_le32( RDES0_STATUS_OWN | RDES0_STATUS_SQL | RDES0_STATUS_FS | RDES0_STATUS_LS );
	}

	/* Setup TX ring. TX buffers descriptors will be filled in as needed */
	for (i = 0; i < TX_RING_SIZE; i++) {
		desc = &priv->tx_ring[i];
		info = &priv->tx_buffers[i];

		info->skb = NULL;
		info->mapping = 0;
		desc->status = 0;
		desc->buffer2 = cpu_to_le32(priv->tx_ring_dma + (i + 1) *
					    sizeof(struct adm8211_desc));
	}
	desc->buffer2 = cpu_to_le32(priv->tx_ring_dma);
	desc->length |= cpu_to_le32(TDES1_CONTROL_TER);
}

#ifdef ADM8211_PROC
static int adm8211_proc_cr_read(char *page, char **start, off_t off,
				int count, int *eof, void *data)
{
	char *p = page;
	struct net_device *dev = data;
	struct adm8211_priv *priv = dev->priv;
	u32 val;

	if (off != 0) {
		*eof = 1;
		return 0;
	}

#define REG(r) \
pci_read_config_dword(priv->pdev, ADM8211_CR_ ## r, &val); \
p += sprintf(p, #r "=%08x\n", val);
	REG(CSC);
	REG(PMR0);
	REG(PMR1);
	REG(LT);
#undef REG

	return (p - page);
}


static int adm8211_proc_csr_read(char *page, char **start, off_t off,
				 int count, int *eof, void *data)
{
	char *p = page;
	struct net_device *dev = data;
	struct adm8211_priv *priv = dev->priv;

	if (off != 0) {
		*eof = 1;
		return 0;
	}

#define REG(r) p += sprintf(p, #r "=%08x\n", ADM8211_CSR_READ(r));
	REG(PAR);
	REG(FRCTL);
	REG(WTDP);
	REG(WRDP);
	REG(RDB);
	REG(TDBD);
	REG(TDBP);
	REG(STSR);
	REG(NAR);
	REG(IER);
	REG(LPC);
	REG(WCSR);
	REG(CMDR);
	REG(PMCSR);
	REG(ABDA1);
	REG(BSSID0);
	REG(TOFS0);
	REG(TOFS1);
	REG(GPIO);
	REG(MIBCNT);
#undef REG

	return (p - page);
}


static int adm8211_proc_tx_read(char *page, char **start, off_t off,
				int count, int *eof, void *data)
{
	char *p = page;
	struct net_device *dev = data;
	struct adm8211_priv *priv = dev->priv;
	int i;
	struct adm8211_desc *desc;
	struct adm8211_ring_info *info;
	u32 status;

	if (off != 0) {
		*eof = 1;
		return 0;
	}

	for (i = 0; i < TX_RING_SIZE; i++) {
		desc = &priv->tx_ring[i];
		info = &priv->tx_buffers[i];
		status = le32_to_cpu(desc->status);
		p += sprintf(p, "%2d: %08x %08x %08x %08x skb=%p "
			     "mapping=%x%s%s\n", i, status,
			     le32_to_cpu(desc->length),
			     le32_to_cpu(desc->buffer1),
			     le32_to_cpu(desc->buffer2),
			     info->skb, info->mapping,
			     status & TDES0_CONTROL_OWN ? " [OWN]" : "",
			     status & TDES0_CONTROL_DONE ? " [DONE]" : "");
	}

	return (p - page);
}


static int adm8211_proc_rx_read(char *page, char **start, off_t off,
				int count, int *eof, void *data)
{
	char *p = page;
	struct net_device *dev = data;
	struct adm8211_priv *priv = dev->priv;
	int i;
	struct adm8211_desc *desc;
	struct adm8211_ring_info *info;
	u32 status;

	if (off != 0) {
		*eof = 1;
		return 0;
	}

	for (i = 0; i < RX_RING_SIZE; i++) {
		desc = &priv->rx_ring[i];
		info = &priv->rx_buffers[i];
		status = le32_to_cpu(desc->status);
		p += sprintf(p, "%2d: %08x %08x %08x %08x skb=%p "
			     "mapping=%x%s\n", i, status,
			     le32_to_cpu(desc->length),
			     le32_to_cpu(desc->buffer1),
			     le32_to_cpu(desc->buffer2),
			     info->skb, info->mapping,
			     status & RDES0_STATUS_OWN ? " [OWN]" : "");
	}

	return (p - page);
}
#endif

static int __devinit adm8211_probe(struct pci_dev *pdev,
				   const struct pci_device_id *id)
{
	int err;
	struct net_device *dev;
	struct adm8211_priv *priv;
	unsigned long mem_addr, mem_len;
	int io_addr, io_len;
	u32 reg;
	int ring_size;

#ifndef MODULE
	static int cardidx;
	if (!cardidx++)
		printk(version);
#endif

	err = pci_enable_device(pdev);
	if (err) {
		printk(KERN_ERR "adm8211: Cannot enable new PCI device\n");
		return err;
	}

	io_addr = pci_resource_start(pdev, 0);
	io_len = pci_resource_len(pdev, 0);
	mem_addr = pci_resource_start(pdev, 1);
	mem_len = pci_resource_len(pdev, 1);
	if (io_len < 256 || mem_len < 1024) {
		printk(KERN_ERR "adm8211: Too short PCI resources\n");
		goto err_disable_pdev;
	}


	/* check signature */
	pci_read_config_dword(pdev, ADM8211_CR_SIG, &reg);
	if (reg != ADM8211_SIG1 && reg != ADM8211_SIG2) {
		printk(KERN_ERR "adm8211: Invalid signature (0x%x)\n", reg);
		goto err_disable_pdev;
	}

	err = pci_request_regions(pdev, "adm8211");
	if (err) {
		printk(KERN_ERR "adm8211: Cannot obtain PCI resources\n");
		goto err_disable_pdev;
	}

	if (pci_set_dma_mask(pdev, 0xffffffff)) {
		printk(KERN_ERR "adm8211: No suitable DMA available\n");
		goto err_free_reg;
	}
	pci_set_consistent_dma_mask(pdev, 0xffffffff);

	pci_set_master(pdev);

	dev = alloc_etherdev(sizeof(struct adm8211_priv));
	if (!dev) {
		printk(KERN_ERR "adm8211: Etherdev alloc failed\n");
		err = -ENOMEM;
		goto err_free_reg;
	}
	priv = dev->priv;
	priv->pdev = pdev;
	priv->eth_header_parse = dev->hard_header_parse;

	spin_lock_init(&priv->lock);

	SET_MODULE_OWNER(dev);

	priv->mem = ioremap(mem_addr, mem_len);
	if (!priv->mem) {
		printk(KERN_ERR "adm8211: Cannot map device memory\n");
		goto err_free_dev;
	}

	priv->io_base = io_addr;

	dev->irq = pdev->irq;

	/* Allocate TX/RX descriptors */
	ring_size = sizeof(struct adm8211_desc) * RX_RING_SIZE +
		sizeof(struct adm8211_desc) * TX_RING_SIZE;
	priv->rx_ring = pci_alloc_consistent(pdev, ring_size,
					     &priv->rx_ring_dma);
	if (!priv->rx_ring) {
		printk(KERN_ERR "adm8211: Cannot allocate TX/RX ring\n");
		goto err_iounmap;
	}
	priv->tx_ring = (struct adm8211_desc *) (priv->rx_ring + RX_RING_SIZE);
	priv->tx_ring_dma = priv->rx_ring_dma +
		sizeof(struct adm8211_desc) * RX_RING_SIZE;

	skb_queue_head_init(&priv->rx_queue);
	priv->rx_tasklet.func = adm8211_rx_tasklet;
	priv->rx_tasklet.data = (unsigned long) dev;

	adm8211_init_rings(dev);

	/* Clear the missed-packet counter. */
	ADM8211_CSR_READ(LPC);

	put_unaligned(ADM8211_CSR_READ(PAR0), (u32 *) dev->dev_addr);
	put_unaligned(ADM8211_CSR_READ(PAR1) & 0xffff,
		      (u16 *) &dev->dev_addr[4]);

	dev->open = adm8211_open;
	dev->stop = adm8211_stop;
	dev->get_stats = adm8211_get_stats;
	dev->set_multicast_list = adm8211_set_rx_mode;
	dev->set_mac_address = adm8211_set_mac_address;
	dev->hard_start_xmit = adm8211_hard_start_xmit;

	dev->get_wireless_stats = adm8211_get_wireless_stats;
	dev->wireless_handlers =
		(struct iw_handler_def *) &adm8211_iw_handler_def;

	priv->iw_mode = IW_MODE_INFRA;

	err = register_netdev(dev);
	if (err) {
		printk(KERN_ERR "adm8211: Cannot register netdevice\n");
		goto err_free_desc;
	}

	pci_set_drvdata(pdev, dev);
	pci_read_config_byte(pdev, ADM8211_CR_CC, &priv->revid);

	adm8211_hw_reset(dev);

	/* FIX: check return value */
	adm8211_read_eeprom(dev);

	printk("%s: hwaddr %02x:%02x:%02x:%02x:%02x:%02x, IRQ %d, Rev 0x%02x, "
	       "PCI ID: 0x%x, mem @0x%lx (len=0x%lx)\n",
	       dev->name, dev->dev_addr[0], dev->dev_addr[1], dev->dev_addr[2],
	       dev->dev_addr[3], dev->dev_addr[4], dev->dev_addr[5],
	       dev->irq, priv->revid, reg, mem_addr, mem_len);
#ifdef ADM8211_PROC
	if (adm8211_proc)
		priv->proc = proc_mkdir(dev->name, adm8211_proc);

	if (priv->proc) {
		create_proc_read_entry("cr", 0, priv->proc,
				       adm8211_proc_cr_read, dev);
		create_proc_read_entry("csr", 0, priv->proc,
				       adm8211_proc_csr_read, dev);
		create_proc_read_entry("tx", 0, priv->proc,
				       adm8211_proc_tx_read, dev);
		create_proc_read_entry("rx", 0, priv->proc,
				       adm8211_proc_rx_read, dev);
	}
#endif
	priv->ieee80211.dev = dev;
	priv->ieee80211.set_channel = adm8211_rf_set_channel;
	priv->ieee80211.set_bssid = adm8211_set_bssid;
	priv->ieee80211.set_ssid = adm8211_set_ssid;
	priv->ieee80211.tx = adm8211_tx;
	priv->ieee80211.auth_update = adm8211_auth_update;
	priv->ieee80211.beacon_update = adm8211_beacon_update;
	priv->ieee80211.link_change = adm8211_link_change;
	priv->ieee80211.set_rate = adm8211_set_rate;
	priv->ieee80211.supp_rates = (u8 *)IEEE80211_B_RATES;
	priv->ieee80211.capab =
		WLAN_CAPABILITY_SHORT_PREAMBLE | WLAN_CAPABILITY_PRIVACY;
	ieee80211_init(&priv->ieee80211);

	priv->frag_thresh = 2346;
	priv->plcp_signal = priv->ieee80211.rate*5;

	return 0;

 err_free_desc:
	pci_free_consistent(pdev, ring_size, priv->rx_ring, priv->rx_ring_dma);

 err_iounmap:
	iounmap(priv->mem);

 err_free_dev:
	kfree(dev);

 err_free_reg:
	pci_release_regions(pdev);

 err_disable_pdev:
	pci_disable_device(pdev);
	pci_set_drvdata(pdev, NULL);
	return err;
}


static void __devexit adm8211_remove(struct pci_dev *pdev)
{
	struct net_device *dev = pci_get_drvdata(pdev);
	struct adm8211_priv *priv;
	struct sk_buff *skb;

	if (!dev)
		return;

	unregister_netdev(dev);

	priv = dev->priv;
	ieee80211_deinit(&priv->ieee80211);
#ifdef ADM8211_PROC
	if (priv->proc) {
		remove_proc_entry("rx", priv->proc);
		remove_proc_entry("tx", priv->proc);
		remove_proc_entry("csr", priv->proc);
		remove_proc_entry("cr", priv->proc);
		remove_proc_entry(dev->name, adm8211_proc);
	}
#endif
	while ((skb = skb_dequeue(&priv->rx_queue)) != NULL)
		dev_kfree_skb(skb);
	iounmap(priv->mem);
	pci_free_consistent(pdev,
			    sizeof(struct adm8211_desc) * RX_RING_SIZE +
			    sizeof(struct adm8211_desc) * TX_RING_SIZE,
			    priv->rx_ring, priv->rx_ring_dma);
	if (priv->eeprom)
		kfree(priv->eeprom);
	pci_release_regions(pdev);
	pci_disable_device(pdev);
	pci_set_drvdata(pdev, NULL);
	free_netdev(dev);
}


#ifdef CONFIG_PM
/* TODO: check if suspend/resume works */
static int adm8211_suspend(struct pci_dev *pdev, u32 state)
{
	struct net_device *dev = pci_get_drvdata(pdev);
	netif_device_detach(dev);

	if (dev->flags & IFF_UP)
		dev->stop(dev);

	pci_disable_device(pdev);
	return 0;
}

static int adm8211_resume(struct pci_dev *pdev)
{
	struct net_device *dev = pci_get_drvdata(pdev);
	pci_enable_device(pdev);

	if (dev->flags & IFF_UP)
		dev->open(dev);

	netif_device_attach(dev);
	return 0;
}
#endif /* CONFIG_PM */


MODULE_DEVICE_TABLE(pci, adm8211_pci_id_table);

/* TODO: enable_wake */
static struct pci_driver adm8211_driver = {
	.name		= "adm8211",
	.id_table	= adm8211_pci_id_table,
	.probe		= adm8211_probe,
	.remove		= __devexit_p(adm8211_remove),
#ifdef CONFIG_PM
	.suspend	= adm8211_suspend,
	.resume		= adm8211_resume,
#endif /* CONFIG_PM */
};



static int __init adm8211_init(void)
{
#ifdef MODULE
	printk(version);
#endif
#ifdef ADM8211_PROC
	if (proc_root_driver)
		adm8211_proc = proc_mkdir("adm8211", proc_root_driver);
#endif
	return pci_module_init(&adm8211_driver);
}


static void __exit adm8211_exit(void)
{
	pci_unregister_driver(&adm8211_driver);
#ifdef ADM8211_PROC
	if (adm8211_proc)
		remove_proc_entry("adm8211", proc_root_driver);
#endif
	printk(KERN_INFO "adm8211: Driver unloaded\n");
}


module_init(adm8211_init);
module_exit(adm8211_exit);
