/*
 * IEEE 802.11 station / management functionality
 *
 * Copyright (c) 2003, Jouni Malinen <jkmaline@cc.hut.fi>
 * Copyright (c) 2004, Michael Wu <flamingice@sourmilk.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation. See README and COPYING for
 * more details.
 */

/* TODO:
 * - implement automatic TX rate control
 * - implement passive scanning
 * - better error reporting
 * - support WEP
 * - support wpa_supplicant from hostap
 * - check supported rates against own supported rates
 */

#include <linux/wireless.h>
#include <linux/if_ether.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
/*#include <linux/crypto.h>*/
#include <linux/crc32.h>
#include <linux/random.h>
#include <asm/types.h>
#include <asm/delay.h>
#include <asm/scatterlist.h>

#include "ieee80211.h"

#define IEEE80211_SCAN_INTERVAL (10 * HZ)
#define IEEE80211_SCAN_LISTEN (HZ / 10)
#define IEEE80211_AUTH_TIMEOUT (1 * HZ)
#define IEEE80211_AUTH_MAX_TRIES 3
#define IEEE80211_ASSOC_TIMEOUT (1 * HZ)
#define IEEE80211_ASSOC_MAX_TRIES 3
#define IEEE80211_MONITORING_INTERVAL (30 * HZ)

static void ieee80211_timer(unsigned long ptr);

ParseRes ieee802_11_parse_elems(u8 *start, size_t len,
				struct ieee802_11_elems *elems)
{
	size_t left = len;
	u8 *pos = start;
	int unknown = 0;

	memset(elems, 0, sizeof(*elems));

	while (left >= 2) {
		u8 id, elen;

		id = *pos++;
		elen = *pos++;
		left -= 2;

		if (elen > left) {
#if 0
			if (net_ratelimit())
				printk(KERN_DEBUG "IEEE 802.11 element parse "
				       "failed (id=%d elen=%d left=%d)\n",
				       id, elen, left);
#endif
			return ParseFailed;
		}

		switch (id) {
		case WLAN_EID_SSID:
			elems->ssid = pos;
			elems->ssid_len = elen;
			break;
		case WLAN_EID_SUPP_RATES:
			elems->supp_rates = pos;
			elems->supp_rates_len = elen;
			break;
		case WLAN_EID_FH_PARAMS:
			elems->fh_params = pos;
			elems->fh_params_len = elen;
			break;
		case WLAN_EID_DS_PARAMS:
			elems->ds_params = pos;
			elems->ds_params_len = elen;
			break;
		case WLAN_EID_CF_PARAMS:
			elems->cf_params = pos;
			elems->cf_params_len = elen;
			break;
		case WLAN_EID_TIM:
			elems->tim = pos;
			elems->tim_len = elen;
			break;
		case WLAN_EID_IBSS_PARAMS:
			elems->ibss_params = pos;
			elems->ibss_params_len = elen;
			break;
		case WLAN_EID_CHALLENGE:
			elems->challenge = pos;
			elems->challenge_len = elen;
			break;
		default:
#if 0
			printk(KERN_DEBUG "IEEE 802.11 element parse ignored "
				      "unknown element (id=%d elen=%d)\n",
				      id, elen);
#endif
			unknown++;
			break;
		}

		left -= elen;
		pos += elen;
	}

	if (left)
		return ParseFailed;

	return unknown ? ParseUnknown : ParseOK;
}


static void ieee80211_dump_ssid(u8 *ssid, size_t ssid_len)
{
	int i;

	for (i = 0; i < ssid_len; i++)
		printk((ssid[i] >= 32 && ssid[i] < 127) ?
		       "%c" : "<%02x>", ssid[i]);
}

static char * ieee80211_status_code(int status)
{
	switch (status) {
	case WLAN_STATUS_SUCCESS:
		return "Success";
	case WLAN_STATUS_UNSPECIFIED_FAILURE:
		return "Unspecified failure";
	case WLAN_STATUS_CAPS_UNSUPPORTED:
		return "Capabilities unsupported";
	case WLAN_STATUS_REASSOC_NO_ASSOC:
		return "";
	case WLAN_STATUS_ASSOC_DENIED_UNSPEC:
		return "";
	case WLAN_STATUS_NOT_SUPPORTED_AUTH_ALG:
		return "Authentication algorithm not supported";
	case WLAN_STATUS_UNKNOWN_AUTH_TRANSACTION:
		return "Unknown authentication transaction";
	case WLAN_STATUS_CHALLENGE_FAIL:
		return "Challenge failed";
	case WLAN_STATUS_AUTH_TIMEOUT:
		return "Authentication timeout";
	case WLAN_STATUS_AP_UNABLE_TO_HANDLE_NEW_STA:
		return "AP unable to handle new STA";
	case WLAN_STATUS_ASSOC_DENIED_RATES:
		return "";

	/* 802.11b */
	case WLAN_STATUS_ASSOC_DENIED_NOSHORT:
		return "";
	case WLAN_STATUS_ASSOC_DENIED_NOPBCC:
		return "";
	case WLAN_STATUS_ASSOC_DENIED_NOAGILITY:
		return "";
	default:
		return "Unknown status code";
	}
}


static void ieee80211_set_associated(struct ieee80211_data *data, int assoc)
{
	if (data->associated == assoc)
		return;

	data->associated = assoc;
	if (data->link_change)
		data->link_change(data->dev, assoc);
}


struct ieee80211_bss * ieee80211_get_bss(struct ieee80211_data *data, u8 *addr)
{
	struct ieee80211_bss *bss;

	bss = data->bss;
	while (bss) {
		if (memcmp(bss->bssid, addr, ETH_ALEN) == 0)
			return bss;
		bss = bss->next;
	}
	return NULL;
}


static int ieee80211_bss_score(struct ieee80211_data *data,
			       struct ieee80211_bss *bss)
{
	int score;

	score = bss->last_rssi;
	score -= 50 * bss->num_auth_fail;
	score -= 50 * bss->num_assoc_fail;
	if (data->pref_channel == bss->channel)
		score += 100;
	return score;
}


static int ieee80211_select_bssid(struct ieee80211_data *data)
{
	struct ieee80211_bss *bss, *prev = NULL, *selected = NULL;
	int best_score = 0, score;

	printk(KERN_DEBUG "%s: select BSSID\n", data->dev->name);

	spin_lock_bh(&data->lock);
	bss = data->bss;

	/* Select first BSS that matches with current configuration */
	while (bss) {
		score = ieee80211_bss_score(data, bss);
		if (score < -250) {
			if (prev) {
				prev->next = bss->next;
				kfree(bss);
				bss = prev->next;
			} else {
				data->bss = bss->next;
				kfree(bss);
				bss = data->bss;
			}
			continue;
		}

#if 1
		printk(KERN_DEBUG "%s: CHAN=%d BSSID=" MACSTR " SSID=",
		       data->dev->name, bss->channel, MAC2STR(bss->bssid));
		ieee80211_dump_ssid(bss->ssid, bss->ssid_len);
		printk(" RSSI=%d num=%lu/%lu score=%d\n", bss->last_rssi,
		       bss->num_beacon, bss->num_proberesp, score);
#endif

		if (data->pref_bssid_set) {
			if (memcmp(bss->bssid, data->pref_bssid, ETH_ALEN) ==
			    0) {
				selected = bss;
				break;
			}
		} else if (data->any_ssid || data->ssid_len == 0 ||
			 (bss->ssid_len == data->ssid_len &&
			  memcmp(bss->ssid, data->ssid, data->ssid_len) == 0))
		{
			if (!selected || score > best_score) {
				best_score = score;
				selected = bss;
			}
		}

		prev = bss;
		bss = bss->next;
	}

	if (selected) {
		if (memcmp(data->bssid, selected->bssid, ETH_ALEN) != 0) {
			printk(KERN_DEBUG "%s: new BSSID " MACSTR "\n",
			       data->dev->name, MAC2STR(selected->bssid));
			ieee80211_set_associated(data, 0);
			data->authenticated = 0;
		}
		memcpy(data->bssid, selected->bssid, ETH_ALEN);
		data->ssid_len = selected->ssid_len;
		memcpy(data->ssid, selected->ssid, data->ssid_len);
		data->channel = selected->channel;
	}

	spin_unlock_bh(&data->lock);

	return selected ? 0 : -1;
}

static void ieee80211_send_auth(struct ieee80211_data *data, u8 *bssid, int status_code)
{
	struct sk_buff *skb;
	struct ieee80211_mgmt *mgmt;
	struct ieee80211_bss *bss;
	u8 *pos;

	if (!data->tx)
		return;

	skb = dev_alloc_skb(sizeof(*mgmt) + 6 + 2 + WLAN_AUTH_CHALLENGE_LEN);
	if (skb == NULL) {
		printk(KERN_DEBUG "%s: failed to allocate buffer for auth "
		       "frame\n", data->dev->name);
		return;
	}

	spin_lock_bh(&data->lock);
	bss = ieee80211_get_bss(data, bssid);

	if (data->mode == IEEE80211_MANAGED && bss->auth_state == 1) {
		/* TODO: error messsage? */
		goto fail;
	}

	mgmt = (struct ieee80211_mgmt *) skb_put(skb, 24 + 6);
	memset(mgmt, 0, 24 + 6);
	mgmt->frame_control = IEEE80211_FC(WLAN_FC_TYPE_MGMT,
					   WLAN_FC_STYPE_AUTH);
	memcpy(mgmt->da, bssid, ETH_ALEN);
	memcpy(mgmt->sa, data->dev->dev_addr, ETH_ALEN);
	memcpy(mgmt->bssid, data->bssid, ETH_ALEN);

	bss->auth_state+=1;
	if (status_code == WLAN_STATUS_NOT_SUPPORTED_AUTH_ALG) {
		mgmt->u.auth.auth_alg = cpu_to_le16(bss->auth_alg);
		mgmt->u.auth.auth_transaction = cpu_to_le16(bss->auth_state);
		mgmt->u.auth.status_code = cpu_to_le16(status_code);
	} else if (bss->auth_alg == WLAN_AUTH_OPEN) {
		mgmt->u.auth.auth_alg = cpu_to_le16(WLAN_AUTH_OPEN);
		mgmt->u.auth.auth_transaction = cpu_to_le16(bss->auth_state);
		mgmt->u.auth.status_code = cpu_to_le16(status_code);

		if (bss->auth_state == 2) {
			bss->auth_state = 0;
			if (status_code == 0)
				bss->authenticated += 1;
		}
	} else if (bss->auth_alg == WLAN_AUTH_SHARED_KEY) {
		mgmt->u.auth.auth_alg = cpu_to_le16(WLAN_AUTH_SHARED_KEY);
		mgmt->u.auth.auth_transaction = cpu_to_le16(bss->auth_state);
		mgmt->u.auth.status_code = cpu_to_le16(status_code);

		switch (bss->auth_state) {
			case 1:
				break;
			case 2: pos = skb_put(skb, 2 + WLAN_AUTH_CHALLENGE_LEN);
				bss->challenge = (u8 *) kmalloc(WLAN_AUTH_CHALLENGE_LEN, GFP_ATOMIC);
				if (!bss->challenge) {
					printk(KERN_DEBUG "%s: failed to allocate buffer "
					       "for challenge\n", data->dev->name);
					bss->auth_state = 0;
					goto fail;
				}
				get_random_bytes(bss->challenge, WLAN_AUTH_CHALLENGE_LEN);
				*pos++ = WLAN_EID_CHALLENGE;
				*pos++ = WLAN_AUTH_CHALLENGE_LEN;
				memcpy(pos, bss->challenge, WLAN_AUTH_CHALLENGE_LEN);
				break;
			case 3: pos = skb_put(skb, 2 + WLAN_AUTH_CHALLENGE_LEN);
				*pos++ = WLAN_EID_CHALLENGE;
				*pos++ = WLAN_AUTH_CHALLENGE_LEN;
				if (bss->challenge)
					memcpy(pos, bss->challenge, WLAN_AUTH_CHALLENGE_LEN);
				else {
					printk(KERN_DEBUG "%s: No challenge to return\n", data->dev->name);
					goto fail;
				}

				ieee80211_data_encode(data, skb);
				break;
			case 4: kfree(bss->challenge);
				bss->auth_state = 0;
				if (status_code == 0)
					bss->authenticated += 1;
				break;
		}
	} else {
		printk(KERN_DEBUG "%s: Unknown authentication algorithm. "
		       "Something is wrong. This shouldn't happen.\n", data->dev->name);
		goto fail;
	}

	spin_unlock_bh(&data->lock);
	data->tx(data->dev, skb);
	return;

fail:
	dev_kfree_skb(skb);
	spin_unlock_bh(&data->lock);
	return;
}


static void ieee80211_authenticate(struct ieee80211_data *data)
{
	data->auth_tries++;
	if (data->auth_tries > IEEE80211_AUTH_MAX_TRIES) {
		struct ieee80211_bss *bss;
		printk(KERN_DEBUG "%s: authentication with AP " MACSTR
		       " CHAN=%d timed out\n",
		       data->dev->name, MAC2STR(data->bssid), data->channel);
		spin_lock_bh(&data->lock);
		bss = ieee80211_get_bss(data, data->bssid);
		if (bss) {
			bss->num_auth_fail++;
			bss->auth_state = 0;
		}
		spin_unlock_bh(&data->lock);
		ieee80211_start_scan(data);
		return;
	}

	data->state = IEEE80211_AUTHENTICATE;
	printk(KERN_DEBUG "%s: authenticate with AP " MACSTR " CHAN=%d\n",
	       data->dev->name, MAC2STR(data->bssid), data->channel);

	if (data->set_bssid)
		data->set_bssid(data->dev, data->bssid);
	if (data->set_ssid)
		data->set_ssid(data->dev, data->ssid, data->ssid_len);
	if (data->set_channel)
		data->set_channel(data->dev, data->channel);

	ieee80211_send_auth(data, data->bssid, 0);

	mod_timer(&data->timer, jiffies + IEEE80211_AUTH_TIMEOUT);
}

#if 0
static void ieee80211_send_assoc_resp(struct ieee80211_data *data, u8 *da, int reassoc, int status)
{
	struct sk_buff *skb;
	struct ieee80211_mgmt *mgmt;
	u8 *pos;

	if (!data->tx)
		return;

	skb = dev_alloc_skb(sizeof(*mgmt) + 2 + 2 + 2 + data->num_supp_rates);
	if (skb == NULL) {
		printk(KERN_DEBUG "%s: failed to allocate buffer for assoc resp "
		       "frame\n", data->dev->name);
		return;
	}

	mgmt = (struct ieee80211_mgmt *) skb_put(skb, 24);
	memset(mgmt, 0, 24);
	memcpy(mgmt->da, da, ETH_ALEN);
	memcpy(mgmt->sa, data->dev->dev_addr, ETH_ALEN);
	memcpy(mgmt->bssid, data->bssid, ETH_ALEN);

	if (reassoc) {
		mgmt->frame_control = IEEE80211_FC(WLAN_FC_TYPE_MGMT,
						   WLAN_FC_STYPE_REASSOC_RESP);
		mgmt->u.reassoc_resp.status_code = cpu_to_le16(status);
		mgmt->u.reassoc_resp.aid = ;
	} else {
		mgmt->frame_control = IEEE80211_FC(WLAN_FC_TYPE_MGMT,
						   WLAN_FC_STYPE_ASSOC_RESP);
		mgmt->u.assoc_resp.status_code = cpu_to_le16(status);
		mgmt->u.assoc_resp.aid = ;
	}

	pos = skb_put(skb, 2 + data->num_supp_rates);
	/* All supported rates */
	*pos++ = WLAN_EID_SUPP_RATES;
	*pos++ = data->num_supp_rates;
	while (i++ < data->num_supp_rates)
		*pos++ = data->supp_rates[i-1];

	data->tx(data->dev, skb);
}
#endif
static void ieee80211_send_assoc_req(struct ieee80211_data *data)
{
	struct sk_buff *skb;
	struct ieee80211_mgmt *mgmt;
	u8 *pos;
	int i = 0;

	if (!data->tx)
		return;

	skb = dev_alloc_skb(sizeof(*mgmt) + 2 + data->ssid_len + 2 + data->num_supp_rates);
	if (skb == NULL) {
		printk(KERN_DEBUG "%s: failed to allocate buffer for assoc req "
		       "frame\n", data->dev->name);
		return;
	}

	mgmt = (struct ieee80211_mgmt *) skb_put(skb, 24);
	memset(mgmt, 0, 24);
	memcpy(mgmt->da, data->bssid, ETH_ALEN);
	memcpy(mgmt->sa, data->dev->dev_addr, ETH_ALEN);
	memcpy(mgmt->bssid, data->bssid, ETH_ALEN);

	/* TODO: check how prev_bssid_set is suppose to work */
	if (data->prev_bssid_set) {
		skb_put(skb, 10);
		mgmt->frame_control = IEEE80211_FC(WLAN_FC_TYPE_MGMT,
						   WLAN_FC_STYPE_REASSOC_REQ);
		mgmt->u.reassoc_req.capab_info = cpu_to_le16(data->capab);
		mgmt->u.reassoc_req.listen_interval =
			cpu_to_le16(data->listen_interval);
		memcpy(mgmt->u.reassoc_req.current_ap, data->prev_bssid,
		       ETH_ALEN);
	} else {
		skb_put(skb, 4);
		mgmt->frame_control = IEEE80211_FC(WLAN_FC_TYPE_MGMT,
						   WLAN_FC_STYPE_ASSOC_REQ);
		mgmt->u.assoc_req.capab_info = cpu_to_le16(data->capab);
		mgmt->u.assoc_req.listen_interval =
			cpu_to_le16(data->listen_interval);
	}

	/* SSID */
	pos = skb_put(skb, 2 + data->ssid_len);
	*pos++ = WLAN_EID_SSID;
	*pos++ = data->ssid_len;
	memcpy(pos, data->ssid, data->ssid_len);

	pos = skb_put(skb, 2 + data->num_supp_rates);
	/* All supported rates */
	*pos++ = WLAN_EID_SUPP_RATES;
	*pos++ = data->num_supp_rates;
	while (i++ < data->num_supp_rates)
		*pos++ = data->supp_rates[i-1];

	data->tx(data->dev, skb);
}


static void ieee80211_associate(struct ieee80211_data *data)
{
	data->assoc_tries++;
	if (data->assoc_tries > IEEE80211_ASSOC_MAX_TRIES) {
		struct ieee80211_bss *bss;
		printk(KERN_DEBUG "%s: association with AP " MACSTR
		       " CHAN=%d timed out\n",
		       data->dev->name, MAC2STR(data->bssid), data->channel);
		spin_lock_bh(&data->lock);
		bss = ieee80211_get_bss(data, data->bssid);
		if (bss) {
			bss->num_assoc_fail++;
			if (bss->num_assoc_fail > 6)
				ieee80211_stop(data);
		}
		spin_unlock_bh(&data->lock);
		ieee80211_start_scan(data);
		return;
	}

	data->state = IEEE80211_ASSOCIATE;
	printk(KERN_DEBUG "%s: associate with AP " MACSTR "\n",
	       data->dev->name, MAC2STR(data->bssid));

	ieee80211_send_assoc_req(data);

	mod_timer(&data->timer, jiffies + IEEE80211_ASSOC_TIMEOUT);
}


static void ieee80211_associated(struct ieee80211_data *data)
{
	/* TODO: start monitoring current AP signal quality and number of
	 * missed beacons. Scan other channels every now and then and search
	 * for better APs. */
	/* TODO: remove expired BSSes */

	if (data->state != IEEE80211_ASSOCIATED) {
		data->state = IEEE80211_ASSOCIATED;
		netif_wake_queue(data->dev);
	}

	mod_timer(&data->timer, jiffies + IEEE80211_MONITORING_INTERVAL);
}

static void ieee80211_send_probe_req(struct ieee80211_data *data)
{
	struct sk_buff *skb;
	struct ieee80211_mgmt *mgmt;
	u8 *pos;
	int i = 0;

	if (!data->tx)
		return;

	skb = dev_alloc_skb(sizeof(*mgmt) + 2 + 2 + data->num_supp_rates);
	if (skb == NULL) {
		printk(KERN_DEBUG "%s: failed to allocate buffer for probe "
		       "request\n", data->dev->name);
		return;
	}

	mgmt = (struct ieee80211_mgmt *) skb_put(skb, 24 + 8);
	memset(mgmt, 0, 24 + 4 + data->num_supp_rates);
	mgmt->frame_control = IEEE80211_FC(WLAN_FC_TYPE_MGMT,
					   WLAN_FC_STYPE_PROBE_REQ);
	memset(mgmt->da, 0xff, ETH_ALEN);
	memcpy(mgmt->sa, data->dev->dev_addr, ETH_ALEN);
	memset(mgmt->bssid, 0xff, ETH_ALEN);
	pos = mgmt->u.probe_req.variable;

	/* Broadcast SSID */
	*pos++ = WLAN_EID_SSID;
	*pos++ = 0;

	/* All supported rates */
	*pos++ = WLAN_EID_SUPP_RATES;
	*pos++ = data->num_supp_rates;
	while (i++ < data->num_supp_rates)
		*pos++ = data->supp_rates[i-1];

	data->tx(data->dev, skb);
}


static void ieee80211_scan(struct ieee80211_data *data)
{
	data->state = IEEE80211_SCAN;
	if (data->channel == 0)
		data->channel = data->chan_range.min;
	else
		data->channel++;

	if (data->channel > data->chan_range.max) {
		if (ieee80211_select_bssid(data) == 0) {
			if (data->mode == IEEE80211_ADHOC)
				memcpy(data->bssid, data->dev->dev_addr, ETH_ALEN);
			ieee80211_authenticate(data);
			return;
		} else {
			/* HACK: remove all BSS and start over */
			ieee80211_stop(data);

			data->channel = 0;
			mod_timer(&data->timer,
				  jiffies + IEEE80211_SCAN_INTERVAL);
			return;
		}
	}

	if (data->set_channel)
		data->set_channel(data->dev, data->channel);

	ieee80211_send_probe_req(data);

	mod_timer(&data->timer, jiffies + IEEE80211_SCAN_LISTEN);
}


void ieee80211_start_scan(struct ieee80211_data *data)
{
	/* Set BSSID to ff:ff:ff:ff:ff:ff for active scanning */
	memset(data->bssid, 0xff, ETH_ALEN);
	data->channel = 0;
	ieee80211_set_associated(data, 0);
	data->authenticated = 0;
	data->auth_tries = data->assoc_tries = 0;

	if (data->set_bssid)
		data->set_bssid(data->dev, data->bssid);
#if 0
	/* doesn't work when called inside an interrupt */
	ieee80211_scan(data);
#else
	data->state = IEEE80211_SCAN;
	mod_timer(&data->timer, jiffies + IEEE80211_SCAN_LISTEN);
#endif
}

void ieee80211_start(struct ieee80211_data *data)
{
	if (data->mode == IEEE80211_MANAGED || data->mode == IEEE80211_ADHOC)
		ieee80211_start_scan(data);
	else
		ieee80211_associated(data);
}

void ieee80211_init(struct ieee80211_data *data)
{
	init_timer(&data->timer);
	data->timer.data = (unsigned long) data;
	data->timer.function = ieee80211_timer;

	spin_lock_init(&data->lock);

	data->listen_interval = 100;
	data->mode = IEEE80211_MANAGED;
	data->any_ssid = 1;
	data->auto_rate = 1;

	if (!data->supp_rates)
		data->supp_rates = (u8 *)IEEE80211_B_RATES;

	data->num_supp_rates=strlen(data->supp_rates);
	data->rate = data->supp_rates[data->num_supp_rates-1];

	get_random_bytes(&data->wep_data[0].iv, 4);
	get_random_bytes(&data->wep_data[1].iv, 4);
	get_random_bytes(&data->wep_data[2].iv, 4);
	get_random_bytes(&data->wep_data[3].iv, 4);
#if 0
	data->tfm = crypto_alloc_tfm("arc4", 0);
	if (!data->tfm)
		printk(KERN_WARNING "%s: arc4 cipher not loaded. "
		       "Software WEP will not work.\n", data->dev->name);
#endif
}

void ieee80211_deinit(struct ieee80211_data *data)
{
	/* nothing for now.. */
	return;
}

void ieee80211_stop(struct ieee80211_data *data)
{
	struct ieee80211_bss *bss, *prev;

	del_timer_sync(&data->timer);

	spin_lock_bh(&data->lock);
	bss = data->bss;
	data->bss = NULL;
	while (bss) {
		prev = bss;
		bss = bss->next;
		kfree(prev);
	}
	spin_unlock_bh(&data->lock);
}


static void ieee80211_add_bss(struct ieee80211_data *data,
			      u8 *bssid, u8 *ssid, size_t ssid_len,
			      int channel, u16 capability, struct ieee80211_mgmt *mgmt,
			      int rssi, int rate, int beacon, int invalid,
			      u8 *supp_rates, size_t supp_rates_len)
{
	struct ieee80211_bss *bss;

	spin_lock_bh(&data->lock);

	bss = data->bss;

	while (bss != NULL &&
	       (memcmp(bss->bssid, bssid, ETH_ALEN) != 0 ||
		bss->ssid_len != ssid_len ||
		memcmp(bss->ssid, ssid, ssid_len) != 0))
		bss = bss->next;

	if (bss == NULL) {
/*		printk(KERN_DEBUG "%s: new BSS - CHAN=%d BSSID=" MACSTR
		       " SSID=",
		       data->dev->name, channel, MAC2STR(bssid));
		ieee80211_dump_ssid(ssid, ssid_len);
		printk("\n");*/

		bss = (struct ieee80211_bss *)
			kmalloc(sizeof(*bss), GFP_ATOMIC);
		if (bss == NULL) {
			spin_unlock_bh(&data->lock);
			printk(KERN_DEBUG "%s: failed to allocate BSS info "
			       "structure\n", data->dev->name);
			return;
		}
		memset(bss, 0, sizeof(*bss));
		memcpy(bss->bssid, bssid, ETH_ALEN);
		bss->last_seq = ~0;
		bss->next = data->bss;
		data->bss = bss;
	}

	if (beacon)
		bss->num_beacon++;
	else
		bss->num_proberesp++;

	memcpy(bss->ssid, ssid, ssid_len);
	bss->ssid_len = ssid_len;
	memset(bss->supp_rates, 0, IEEE80211_MAX_SUPP_RATES);
	if (supp_rates) {
		memcpy(bss->supp_rates, supp_rates,
		       supp_rates_len > IEEE80211_MAX_SUPP_RATES ?
		       IEEE80211_MAX_SUPP_RATES : supp_rates_len);
	}
	bss->channel = channel;
	bss->interval = le16_to_cpu(mgmt->u.beacon.beacon_int);
	bss->timestamp = le64_to_cpu(*(u64 *)(mgmt->u.beacon.timestamp));
	bss->capability = capability;
	bss->last_rx = jiffies;
	if (rssi)
		bss->last_rssi = rssi;
	bss->last_rate = rate;
	bss->invalid_frame = invalid;

	if (capability & WLAN_CAPABILITY_PRIVACY)
		bss->auth_alg = 1;

	/* automatically authenticate w/ bss if we're in adhoc */
	if (!bss->authenticated && data->mode == IEEE80211_ADHOC &&
	    !memcmp(data->ssid, ssid, ssid_len)) {
		spin_unlock_bh(&data->lock);
		ieee80211_send_auth(data, bssid, 0);
	} else
		spin_unlock_bh(&data->lock);
}


/* Process beacon and probe response frames in common function since they have
 * almost identical contents. */
static void ieee80211_rx_mgmt_bss_info(struct ieee80211_data *data,
				       struct ieee80211_mgmt *mgmt,
				       size_t len,
				       struct ieee80211_rx_status *rx_status,
				       int beacon)
{
	struct ieee802_11_elems elems;
	size_t baselen;
	int channel, invalid = 0;

#if 0
	printk(KERN_DEBUG "%s: %s from " MACSTR " rate=%d rssi=%d\n",
	       data->dev->name, beacon ? "Beacon" : "ProbeResp",
	       MAC2STR(mgmt->sa), rx_status->rate, rx_status->rssi);
#endif

	baselen = (u8 *) mgmt->u.beacon.variable - (u8 *) mgmt;
	if (baselen > len) {
		printk(KERN_DEBUG "%s: Beacon/ProbeResp underflow %d > %d\n",
		       data->dev->name, baselen, len);
		return;
	}

	if (ieee802_11_parse_elems(mgmt->u.beacon.variable, len - baselen,
				   &elems) == ParseFailed)
		invalid = 1;

	if (elems.ssid == NULL)
		return;

	if (elems.ds_params && elems.ds_params_len == 1)
		channel = elems.ds_params[0];
	else
		channel = data->channel;

	ieee80211_add_bss(data, mgmt->sa, elems.ssid, elems.ssid_len, channel,
			  le16_to_cpu(mgmt->u.beacon.capab_info), mgmt,
			  rx_status->rssi, rx_status->rate, beacon, invalid,
			  elems.supp_rates, elems.supp_rates_len);
}

#if 0
static void ieee80211_rx_mgmt_probe_req(struct ieee80211_data *data,
					struct ieee80211_mgmt *mgmt,
					size_t len,
					struct ieee80211_rx_status *rx_status)
{
	if (data->mode != IEEE80211_ADHOC && data->mode != IEEE80211_AP)
		return;

	/* FIX: Software probe response not implemented (not needed) */
}
#endif
static void ieee80211_rx_mgmt_probe_resp(struct ieee80211_data *data,
					 struct ieee80211_mgmt *mgmt,
					 size_t len,
					 struct ieee80211_rx_status *rx_status)
{
	ieee80211_rx_mgmt_bss_info(data, mgmt, len, rx_status, 0);
}


static void ieee80211_rx_mgmt_beacon(struct ieee80211_data *data,
				     struct ieee80211_mgmt *mgmt,
				     size_t len,
				     struct ieee80211_rx_status *rx_status)
{
	ieee80211_rx_mgmt_bss_info(data, mgmt, len, rx_status, 1);
	if (data->beacon_update)
		data->beacon_update(data->dev, mgmt->duration);
}

/* TODO: filter auth_transaction and make sure it makes sense for the mode we're in */
static void ieee80211_rx_mgmt_auth(struct ieee80211_data *data,
				   struct ieee80211_mgmt *mgmt,
				   size_t len,
				   struct ieee80211_rx_status *rx_status, struct sk_buff *skb)
{
	u16 auth_alg, auth_transaction, status_code;
	struct ieee802_11_elems elems;
	struct ieee80211_bss *bss;
	size_t baselen;

	if (len < 24 + 6) {
		printk(KERN_DEBUG "%s: too short (%d) authentication frame "
		       "received from " MACSTR " - ignored\n",
		       data->dev->name, len, MAC2STR(mgmt->sa));
		return;
	}

	if (data->mode == IEEE80211_MANAGED) {
	if (data->state != IEEE80211_AUTHENTICATE) {
		printk(KERN_DEBUG "%s: authentication frame received from "
		       MACSTR ", but not in authenticate state - ignored\n",
		       data->dev->name, MAC2STR(mgmt->sa));
		return;
	}
	if (memcmp(data->bssid, mgmt->sa, ETH_ALEN) != 0) {
		printk(KERN_DEBUG "%s: authentication frame received from "
		       "unknown AP (SA=" MACSTR " BSSID=" MACSTR ") - "
		       "ignored\n", data->dev->name, MAC2STR(mgmt->sa),
		       MAC2STR(mgmt->bssid));
		return;
	}
	}

	baselen = (u8 *) mgmt->u.auth.variable - (u8 *) mgmt;
	if (baselen > len) {
		printk(KERN_DEBUG "%s: Auth underflow %d > %d\n",
		       data->dev->name, baselen, len);
		return;
	}

	if (ieee802_11_parse_elems(mgmt->u.auth.variable, len - baselen,
				   &elems) == ParseFailed) {
		printk(KERN_DEBUG "%s: Parse failed.\n", data->dev->name);
		return;
	}

	auth_alg = le16_to_cpu(mgmt->u.auth.auth_alg);
	auth_transaction = le16_to_cpu(mgmt->u.auth.auth_transaction);
	status_code = le16_to_cpu(mgmt->u.auth.status_code);

	printk(KERN_DEBUG "%s: RX authentication from " MACSTR " (alg=%d "
	       "transaction=%d status=%d %s)\n",
	       data->dev->name, MAC2STR(mgmt->sa), auth_alg,
	       auth_transaction, status_code,
	       ieee80211_status_code(status_code));

	if (auth_transaction & 0x1 && status_code != WLAN_STATUS_SUCCESS) {
		/* FIX: message doesn't apply in adhoc */
		printk(KERN_DEBUG "%s: AP denied authentication (code=%d)\n",
		       data->dev->name, status_code);
		spin_lock_bh(&data->lock);
		bss = ieee80211_get_bss(data, mgmt->sa);
		if (bss)
			bss->num_auth_fail++;
		spin_unlock_bh(&data->lock);
		return;
	}

	spin_lock_bh(&data->lock);
	bss = ieee80211_get_bss(data, mgmt->sa);
	if (!bss) {
		spin_unlock_bh(&data->lock);
		return;
	}

	if (auth_transaction != bss->auth_state + 1) {
		printk(KERN_DEBUG "%s: unexpected authentication "
		       "transaction number: %d\n", data->dev->name, auth_transaction);
		bss->auth_state = 0;
		spin_unlock_bh(&data->lock);
		return;
	}
	bss->auth_alg = auth_alg;
	bss->auth_state = auth_transaction;
	spin_unlock_bh(&data->lock);

	if (data->wep_enabled & 0x2 &&
	   (auth_alg != data->auth_algorithm || data->auth_algorithm > 1) &&
	   auth_transaction == 1) {
		ieee80211_send_auth(data, mgmt->sa, WLAN_STATUS_NOT_SUPPORTED_AUTH_ALG);
		return;
	}

	switch (auth_alg) {
		case WLAN_AUTH_OPEN:
			if (auth_transaction == 1)
				ieee80211_send_auth(data, mgmt->sa, 0);
			else if (auth_transaction == 2)
				data->authenticated = 1;
			else
				printk(KERN_DEBUG "%s: invalid authentication "
				       "transaction number: %d\n",
				       data->dev->name, auth_transaction);
			break;
		case WLAN_AUTH_SHARED_KEY:
			switch (auth_transaction) {
				case 1:
					ieee80211_send_auth(data, mgmt->sa, 0);
					break;
				case 2:
					/* TODO: error messages? */
					if (elems.challenge_len != WLAN_AUTH_CHALLENGE_LEN) {
						printk(KERN_DEBUG "%s: invalid challenge length: %d\n",
						       data->dev->name, elems.challenge_len);
						return;
					}
					spin_lock_bh(&data->lock);
					bss = ieee80211_get_bss(data, mgmt->sa);
					if (!bss) {
						spin_unlock_bh(&data->lock);
						return;
					}
					if (bss->challenge)
						kfree(bss->challenge);
					bss->challenge = (u8 *) kmalloc(elems.challenge_len, GFP_ATOMIC);
					memcpy(bss->challenge, elems.challenge, elems.challenge_len);
					spin_unlock_bh(&data->lock);
					ieee80211_send_auth(data, mgmt->sa, 0);
					break;
				case 3:
					if (mgmt->frame_control & WLAN_FC_ISWEP)
						skb = ieee80211_data_decode(data, skb);
					else {
						ieee80211_send_auth(data, mgmt->sa, WLAN_STATUS_CHALLENGE_FAIL);
						return;
					}
					if (!skb) {
						ieee80211_send_auth(data, mgmt->sa, WLAN_STATUS_CHALLENGE_FAIL);
						return;
					}
					if (ieee802_11_parse_elems(mgmt->u.beacon.variable,
						len - baselen, &elems) == ParseFailed)
						return;
					if (!bss->challenge || memcmp(elems.challenge,
						bss->challenge, WLAN_AUTH_CHALLENGE_LEN)) {
						ieee80211_send_auth(data, mgmt->sa, WLAN_STATUS_CHALLENGE_FAIL);
						return;
					}
					ieee80211_send_auth(data, mgmt->sa, 0);
					break;
				case 4:
					data->authenticated = 1;
					break;
				default:
					printk(KERN_DEBUG "%s: invalid authentication"
					       "transaction number: %d\n",
					       data->dev->name, auth_transaction);
					break;
			}
			break;
	}

	if (data->authenticated) {
	        spin_lock_bh(&data->lock);
	        bss = ieee80211_get_bss(data, mgmt->sa);
		bss->auth_state = 0;
		spin_unlock_bh(&data->lock);

		printk(KERN_DEBUG "%s: authenticated\n", data->dev->name);
		if (data->auth_update)
			data->auth_update(data->dev);

		if (data->mode == IEEE80211_MANAGED)
			ieee80211_associate(data);
		else if (data->mode == IEEE80211_ADHOC)
			ieee80211_associated(data);
	}

}

static void ieee80211_rx_mgmt_deauth(struct ieee80211_data *data,
				     struct ieee80211_mgmt *mgmt,
				     size_t len,
				     struct ieee80211_rx_status *rx_status)
{
	struct ieee80211_bss *bss;
	u16 reason_code;

	if (len < 24 + 2) {
		printk(KERN_DEBUG "%s: too short (%d) deauthentication frame "
		       "received from " MACSTR " - ignored\n",
		       data->dev->name, len, MAC2STR(mgmt->sa));
		return;
	}

	if (memcmp(data->bssid, mgmt->sa, ETH_ALEN) != 0) {
		printk(KERN_DEBUG "%s: deauthentication frame received from "
		       "unknown AP (SA=" MACSTR " BSSID=" MACSTR ") - "
		       "ignored\n", data->dev->name, MAC2STR(mgmt->sa),
		       MAC2STR(mgmt->bssid));
		return;
	}

	reason_code = le16_to_cpu(mgmt->u.deauth.reason_code);

	printk(KERN_DEBUG "%s: RX deauthentication from " MACSTR
	       " (reason=%d)\n",
	       data->dev->name, MAC2STR(mgmt->sa), reason_code);

	spin_lock_bh(&data->lock);
	bss = ieee80211_get_bss(data, mgmt->sa);
	if (bss) {
		bss->num_auth_fail++;
		bss->authenticated = 0;
	}
	spin_unlock_bh(&data->lock);

	if (data->mode == IEEE80211_MANAGED) {
	if (data->authenticated) {
		printk(KERN_DEBUG "%s: deauthenticated\n", data->dev->name);
	}
	ieee80211_set_associated(data, 0);
	data->authenticated = 0;

	ieee80211_scan(data);
	}
}


static void ieee80211_rx_mgmt_disassoc(struct ieee80211_data *data,
				       struct ieee80211_mgmt *mgmt,
				       size_t len,
				       struct ieee80211_rx_status *rx_status)
{
	struct ieee80211_bss *bss;
	u16 reason_code;

	if (len < 24 + 2) {
		printk(KERN_DEBUG "%s: too short (%d) disassociation frame "
		       "received from " MACSTR " - ignored\n",
		       data->dev->name, len, MAC2STR(mgmt->sa));
		return;
	}

	if (memcmp(data->bssid, mgmt->sa, ETH_ALEN) != 0) {
		printk(KERN_DEBUG "%s: disassociation frame received from "
		       "unknown AP (SA=" MACSTR " BSSID=" MACSTR ") - "
		       "ignored\n", data->dev->name, MAC2STR(mgmt->sa),
		       MAC2STR(mgmt->bssid));
		return;
	}

	reason_code = le16_to_cpu(mgmt->u.disassoc.reason_code);

	printk(KERN_DEBUG "%s: RX disassociation from " MACSTR
	       " (reason=%d)\n",
	       data->dev->name, MAC2STR(mgmt->sa), reason_code);

	spin_lock_bh(&data->lock);
	bss = ieee80211_get_bss(data, mgmt->sa);
	if (bss)
		bss->num_assoc_fail++;
	spin_unlock_bh(&data->lock);

	if (data->associated)
		printk(KERN_DEBUG "%s: disassociated\n", data->dev->name);
	ieee80211_set_associated(data, 0);

	ieee80211_scan(data);
}

static void ieee80211_rx_mgmt_assoc_req(struct ieee80211_data *data,
					struct ieee80211_mgmt *mgmt,
					size_t len,
					struct ieee80211_rx_status *rx_status,
					int reassoc)
{
	if (reassoc && data->mode == IEEE80211_MANAGED)
		ieee80211_associate(data);

	return;
}

static void ieee80211_rx_mgmt_assoc_resp(struct ieee80211_data *data,
					 struct ieee80211_mgmt *mgmt,
					 size_t len,
					 struct ieee80211_rx_status *rx_status,
					 int reassoc)
{
	struct ieee802_11_elems elems;
	struct ieee80211_bss *bss;
	size_t baselen;
	u16 capab_info, status_code, aid;

	/* AssocResp and ReassocResp have identical structure, so process both
	 * of them in this function. */

	if (data->state != IEEE80211_ASSOCIATE) {
		printk(KERN_DEBUG "%s: association frame received from "
		       MACSTR ", but not in associate state - ignored\n",
		       data->dev->name, MAC2STR(mgmt->sa));
		return;
	}

	if (len < 24 + 6) {
		printk(KERN_DEBUG "%s: too short (%d) association frame "
		       "received from " MACSTR " - ignored\n",
		       data->dev->name, len, MAC2STR(mgmt->sa));
		return;
	}

	if (memcmp(data->bssid, mgmt->sa, ETH_ALEN) != 0) {
		printk(KERN_DEBUG "%s: association frame received from "
		       "unknown AP (SA=" MACSTR " BSSID=" MACSTR ") - "
		       "ignored\n", data->dev->name, MAC2STR(mgmt->sa),
		       MAC2STR(mgmt->bssid));
		return;
	}

	capab_info = le16_to_cpu(mgmt->u.assoc_resp.capab_info);
	status_code = le16_to_cpu(mgmt->u.assoc_resp.status_code);
	aid = le16_to_cpu(mgmt->u.assoc_resp.aid);
	if ((aid & (BIT(15) | BIT(14))) != (BIT(15) | BIT(14)))
		printk(KERN_DEBUG "%s: invalid aid value %d; bits 15:14 not "
		       "set\n", data->dev->name, aid);
	aid &= ~(BIT(15) | BIT(14));

	printk(KERN_DEBUG "%s: RX %sssocResp from " MACSTR " (capab=0x%x "
	       "aid=%d status=%d %s)\n",
	       data->dev->name, reassoc ? "Rea" : "A", MAC2STR(mgmt->sa),
	       capab_info, aid, status_code,
	       ieee80211_status_code(status_code));

	spin_lock_bh(&data->lock);
	bss = ieee80211_get_bss(data, mgmt->sa);
	if (status_code != WLAN_STATUS_SUCCESS) {
		printk(KERN_DEBUG "%s: AP denied association (code=%d)\n",
		       data->dev->name, status_code);
		if (bss)
			bss->num_assoc_fail++;
		spin_unlock_bh(&data->lock);
		return;
	}
	if (bss) {
		bss->num_auth_fail = bss->num_assoc_fail = 0;
	}

	baselen = (u8 *) mgmt->u.assoc_resp.variable - (u8 *) mgmt;
        if (ieee802_11_parse_elems(mgmt->u.assoc_resp.variable, len - baselen, &elems) == ParseOK
	    && elems.supp_rates) {
		memset(bss->supp_rates, 0, IEEE80211_MAX_SUPP_RATES);
		memcpy(bss->supp_rates, elems.supp_rates,
		       elems.supp_rates_len > IEEE80211_MAX_SUPP_RATES ?
		       IEEE80211_MAX_SUPP_RATES : elems.supp_rates_len);
	}

	spin_unlock_bh(&data->lock);

	printk(KERN_DEBUG "%s: associated\n", data->dev->name);
	data->aid = aid;
	data->ap_capab = capab_info;
	ieee80211_set_associated(data, 1);
	ieee80211_associated(data);
}


void ieee80211_rx_mgmt(struct ieee80211_data *data, struct sk_buff *skb,
		       struct ieee80211_rx_status *rx_status)
{
	struct ieee80211_mgmt *mgmt;
	u16 fc;

	if (skb->len < 24)
		goto fail;

	mgmt = (struct ieee80211_mgmt *) skb->data;
	fc = le16_to_cpu(mgmt->frame_control);

	switch (WLAN_FC_GET_STYPE(fc)) {
	case WLAN_FC_STYPE_PROBE_REQ:
#if 0
		ieee80211_rx_mgmt_probe_req(data, mgmt, skb->len, rx_status);
#endif
		break;
	case WLAN_FC_STYPE_PROBE_RESP:
		ieee80211_rx_mgmt_probe_resp(data, mgmt, skb->len, rx_status);
		break;
	case WLAN_FC_STYPE_BEACON:
		ieee80211_rx_mgmt_beacon(data, mgmt, skb->len, rx_status);
		break;
	case WLAN_FC_STYPE_AUTH:
		ieee80211_rx_mgmt_auth(data, mgmt, skb->len, rx_status, skb);
		break;
	case WLAN_FC_STYPE_ASSOC_REQ:
		ieee80211_rx_mgmt_assoc_req(data, mgmt, skb->len, rx_status,
					    0);
		break;
	case WLAN_FC_STYPE_ASSOC_RESP:
		ieee80211_rx_mgmt_assoc_resp(data, mgmt, skb->len, rx_status,
					     0);
		break;
	case WLAN_FC_STYPE_REASSOC_REQ:
		ieee80211_rx_mgmt_assoc_req(data, mgmt, skb->len, rx_status,
					    1);
		break;
	case WLAN_FC_STYPE_REASSOC_RESP:
		ieee80211_rx_mgmt_assoc_resp(data, mgmt, skb->len, rx_status,
					     1);
		break;
	case WLAN_FC_STYPE_DEAUTH:
		ieee80211_rx_mgmt_deauth(data, mgmt, skb->len, rx_status);
		break;
	case WLAN_FC_STYPE_DISASSOC:
		ieee80211_rx_mgmt_disassoc(data, mgmt, skb->len, rx_status);
		break;
	default:
		printk(KERN_DEBUG "%s: received unknown management frame - "
		       "stype=%d\n", data->dev->name, WLAN_FC_GET_STYPE(fc));
		break;
	}

 fail:
	if (skb)
		dev_kfree_skb(skb);
}


static void ieee80211_timer(unsigned long ptr)
{
	struct ieee80211_data *data = (struct ieee80211_data *) ptr;

	switch (data->state) {
	case IEEE80211_SCAN:
		ieee80211_scan(data);
		break;
	case IEEE80211_AUTHENTICATE:
		ieee80211_authenticate(data);
		break;
	case IEEE80211_ASSOCIATE:
		ieee80211_associate(data);
		break;
	case IEEE80211_ASSOCIATED:
		ieee80211_associated(data);
		break;
	default:
		printk(KERN_DEBUG "ieee80211_timer: Unknown state %d\n",
		       data->state);
		break;
	}
}

/* TODO:rename these WEP functions */
struct sk_buff * ieee80211_data_encode(struct ieee80211_data *data,
				       struct sk_buff *skb)
{
	struct ieee80211_hdr *hdr;
	u16 fc;
	int hlen, rc;

	hdr = (struct ieee80211_hdr *) skb->data;
	fc = le16_to_cpu(hdr->frame_control);

	if (!data->wep_enabled)
		return skb;
	else
		fc |= WLAN_FC_ISWEP;

        if ((fc & (WLAN_FC_TODS | WLAN_FC_FROMDS)) ==
                  (WLAN_FC_TODS | WLAN_FC_FROMDS))
                hlen = 24;
        else
                hlen = 30;

	rc = wep_encrypt(skb, hlen, data->default_key,
			 &data->wep_data[data->default_key]);
	if (rc) {
		dev_kfree_skb(skb);
		return NULL;
	} else
		return skb;

	return skb;
}
#if 1
struct sk_buff * ieee80211_data_decode(struct ieee80211_data *data,
				       struct sk_buff *skb)
{
	struct ieee80211_hdr *hdr;
	u16 fc;
	int hlen, index, rc;

	hdr = (struct ieee80211_hdr *) skb->data;
	fc = le16_to_cpu(hdr->frame_control);

	if (!(fc & WLAN_FC_ISWEP))
		return skb;

	hdr->frame_control &= cpu_to_le16(~WLAN_FC_ISWEP);

	if ((fc & (WLAN_FC_TODS | WLAN_FC_FROMDS)) ==
		  (WLAN_FC_TODS | WLAN_FC_FROMDS))
		hlen = 24;
	else
		hlen = 30;

	index = (*(skb->data + hlen + 3) >> 6) & 0x3;

	rc = wep_decrypt(skb, hlen, index, &data->wep_data[index]);
	if (rc) {
		dev_kfree_skb(skb);
		return NULL;
	} else
		return skb;
}
#endif

/* See IEEE 802.1H for LLC/SNAP encapsulation/decapsulation */
/* Ethernet-II snap header (RFC1042 for most EtherTypes) */
static unsigned char rfc1042_header[] =
{ 0xaa, 0xaa, 0x03, 0x00, 0x00, 0x00 };
/* Bridge-Tunnel header (for EtherTypes ETH_P_AARP and ETH_P_IPX) */
static unsigned char bridge_tunnel_header[] =
{ 0xaa, 0xaa, 0x03, 0x00, 0x00, 0xf8 };
/* No encapsulation header if EtherType < 0x600 (=length) */


struct sk_buff * ieee80211_data_encaps(struct ieee80211_data *data,
				       struct sk_buff *skb)
{
	struct net_device *dev = data->dev;
	u16 ethertype;
	u8 dst[ETH_ALEN];
	u8 src[ETH_ALEN];
	struct ieee80211_hdr *hdr;
	int hlen;
	u16 fc;

	/* hold the packets till we're authenticated/associated */
	if ((data->mode == IEEE80211_MANAGED && !data->associated) ||
	    (data->mode == IEEE80211_ADHOC && !data->authenticated)) {
		dev_kfree_skb(skb);
		netif_stop_queue(dev);
		return NULL;
	}

	if (skb->len < ETH_HLEN) {
		dev_kfree_skb(skb);
		return NULL;
	}

	ethertype = (skb->data[12] << 8) | skb->data[13];
	memcpy(dst, skb->data, ETH_ALEN);
	memcpy(src, skb->data + ETH_ALEN, ETH_ALEN);

	skb_pull(skb, 12);
	hlen = 24 + 6;
	if (data->mode == IEEE80211_WDS)
		hlen += ETH_ALEN;

	if (skb_headroom(skb) < hlen) {
		if (pskb_expand_head(skb, hlen, 0, GFP_ATOMIC)) {
			printk(KERN_DEBUG "%s: failed to reallocate skb "
			       "buffer\n", dev->name);
			dev_kfree_skb(skb);
			return NULL;
		}
	}

	memcpy(skb_push(skb, 6),
	       ethertype == ETH_P_AARP || ethertype == ETH_P_IPX ?
	       bridge_tunnel_header : rfc1042_header, 6);

	hdr = (struct ieee80211_hdr *) skb_push(skb, hlen - 6);

	fc = (WLAN_FC_TYPE_DATA << 2) | (WLAN_FC_STYPE_DATA << 4);

	switch (data->mode) {
	case IEEE80211_MANAGED:
		memcpy(hdr->addr1, data->bssid, ETH_ALEN); /* BSSID */
		memcpy(hdr->addr2, src, ETH_ALEN); /* SA */
		memcpy(hdr->addr3, dst, ETH_ALEN); /* DA */
		fc |= WLAN_FC_TODS;
		break;
	case IEEE80211_ADHOC:
		memcpy(hdr->addr1, dst, ETH_ALEN); /* DA */
		memcpy(hdr->addr2, src, ETH_ALEN); /* SA */
		memcpy(hdr->addr3, data->bssid, ETH_ALEN); /* BSSID */
		break;
	case IEEE80211_AP:
		memcpy(hdr->addr1, dst, ETH_ALEN); /* DA */
		memcpy(hdr->addr2, dev->dev_addr, ETH_ALEN); /* BSSID */
		memcpy(hdr->addr3, src, ETH_ALEN); /* SA */
		fc |= WLAN_FC_FROMDS;
		break;
	case IEEE80211_WDS:
		memcpy(hdr->addr1, dst, ETH_ALEN); /* RA */
		memcpy(hdr->addr2, dev->dev_addr, ETH_ALEN); /* TA */
		memcpy(hdr->addr3, dst, ETH_ALEN); /* DA */
		memcpy(hdr->addr4, src, ETH_ALEN); /* SA */
		fc |= WLAN_FC_FROMDS | WLAN_FC_TODS;
		break;
	}

	hdr->frame_control = cpu_to_le16(fc);
	hdr->duration_id = 0;
	hdr->seq_ctrl = 0;

	return ieee80211_data_encode(data, skb);
}


struct sk_buff * ieee80211_data_decaps(struct ieee80211_data *data,
				       struct sk_buff *skb)
{
	u8 dst[ETH_ALEN];
	u8 src[ETH_ALEN];
	struct ieee80211_hdr *hdr;
	int hlen;
	u16 fc;
	u8 *pos;

	if (skb->len < 24)
		goto drop;

	skb = ieee80211_data_decode(data, skb);

	if (!skb)
		return NULL;

	hdr = (struct ieee80211_hdr *) skb->data;
	fc = le16_to_cpu(hdr->frame_control);

	hlen = 24;

	switch (fc & (WLAN_FC_TODS | WLAN_FC_FROMDS)) {
	case WLAN_FC_TODS | WLAN_FC_FROMDS:
		/* RA, TA, DA, SA */
		if (skb->len < 30)
			goto drop;
		memcpy(dst, hdr->addr3, ETH_ALEN);
		memcpy(src, hdr->addr4, ETH_ALEN);
		hlen += 6;
		break;
	case WLAN_FC_TODS:
		/* BSSID, SA, DA */
		memcpy(dst, hdr->addr3, ETH_ALEN);
		memcpy(src, hdr->addr2, ETH_ALEN);
		break;
	case WLAN_FC_FROMDS:
		/* DA, BSSID, SA */
		memcpy(dst, hdr->addr1, ETH_ALEN);
		memcpy(src, hdr->addr3, ETH_ALEN);
		break;
	case 0:
		/* DA, SA, BSSID */
		memcpy(dst, hdr->addr1, ETH_ALEN);
		memcpy(src, hdr->addr2, ETH_ALEN);
		break;
	}

	pos = skb_pull(skb, hlen);
	if (hlen >= 6 &&
	    (memcmp(pos, rfc1042_header, 6) == 0 ||
	     memcmp(pos, bridge_tunnel_header, 6) == 0)) {
		skb_pull(skb, 6);
	}

	memcpy(skb_push(skb, ETH_ALEN), src, ETH_ALEN);
	memcpy(skb_push(skb, ETH_ALEN), dst, ETH_ALEN);

	return skb;

 drop:
	dev_kfree_skb(skb);
	return NULL;
}

int ieee80211_filter_duplicates(struct ieee80211_data *data,
				struct sk_buff *skb, u16 fc)
{
	struct ieee80211_hdr *hdr;
	struct ieee80211_bss *bss;

	hdr = (struct ieee80211_hdr *) skb->data;

	spin_lock_bh(&data->lock);
	bss = ieee80211_get_bss(data, hdr->addr2);

	if (bss) {
		if (fc & WLAN_FC_RETRY && bss->last_seq == hdr->seq_ctrl) {
			spin_unlock_bh(&data->lock);
			return 1;
		}
		bss->last_seq=hdr->seq_ctrl;
	}

	spin_unlock_bh(&data->lock);

	return 0;
}

/* returns a rate compatible w/ the destination, if it can be determined */
int ieee80211_get_rate(struct ieee80211_data *data, u8 *addr)
{
	struct ieee80211_bss *bss;
	int rate;

	spin_lock_bh(&data->lock);
	bss = ieee80211_get_bss(data, addr);

	if (bss) {
		/* TODO: finish */
		rate = data->rate;
	} else {
		rate = data->rate;
	}
	spin_unlock_bh(&data->lock);

	return rate;
}
